###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Common functions for converting types to and from XML files
"""
from __future__ import annotations

from typing import Iterable, Any, cast
import sys
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  #type: ignore[misc]
from lxml import etree
import logging
from masci_tools.io.parsers import fleur_schema
import re

BaseType: TypeAlias = Literal['int', 'switch', 'string', 'float', 'float_expression', 'complex']
ConvertedType: TypeAlias = 'int | float | bool | str | complex'


def convert_to_xml(value: Any | list[Any],
                   schema_dict: fleur_schema.SchemaDict,
                   name: str,
                   text: bool = False,
                   logger: logging.Logger | None = None,
                   list_return: bool = False) -> tuple[str | list[str], bool]:
    """
    Tries to converts a given string to the types specified in the schema_dict.
    First succeeded conversion will be returned

    If no logger is given and a attribute cannot be converted an error is raised

    :param stringattribute: str, Attribute to convert.
    :param schema_dict: Schema dictionary containing all the information
    :param name: name of the attribute or element
    :param text: bool, decides whether to take the definitions for text or attributes
    :param constants: dict, of constants defined in fleur input
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted value of the first successful conversion
    """

    if text:
        if name not in schema_dict['text_types']:
            raise KeyError(f'Unknown text tag: {name}')
        definitions = schema_dict['text_types'][name]
        float_format = '16.13'
    else:
        if name not in schema_dict['attrib_types']:
            raise KeyError(f'Unknown attribute name: {name}')
        definitions = schema_dict['attrib_types'][name]
        float_format = '.10'

    return convert_to_xml_explicit(value,
                                   definitions,
                                   logger=logger,
                                   list_return=list_return,
                                   float_format=float_format)


def convert_from_xml(
    xmlstring: str | list[str],
    schema_dict: fleur_schema.SchemaDict,
    name: str,
    text: bool = False,
    constants: dict[str, float] | None = None,
    logger: logging.Logger | None = None,
    list_return: bool = False
) -> tuple[ConvertedType | list[ConvertedType] | list[ConvertedType | list[ConvertedType]], bool]:
    """
    Tries to converts a given string to the types specified in the schema_dict.
    First succeeded conversion will be returned

    If no logger is given and a attribute cannot be converted an error is raised

    :param stringattribute: str, Attribute to convert.
    :param schema_dict: Schema dictionary containing all the information
    :param name: name of the attribute or element
    :param text: bool, decides whether to take the definitions for text or attributes
    :param constants: dict, of constants defined in fleur input
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted value of the first successful conversion
    """

    if not isinstance(xmlstring, list):
        xmlstring = [xmlstring]

    if text:
        if name not in schema_dict['text_types']:
            raise KeyError(f'Unknown text tag: {name}')
        definitions = schema_dict['text_types'][name]
        xmlstring = [string.strip() for string in xmlstring]
    else:
        if name not in schema_dict['attrib_types']:
            raise KeyError(f'Unknown attribute name: {name}')
        definitions = schema_dict['attrib_types'][name]

    return convert_from_xml_explicit(xmlstring,
                                     definitions,
                                     constants=constants,
                                     logger=logger,
                                     list_return=list_return)


def convert_from_xml_explicit(
    xmlstring: str | list[str],
    definitions: list[fleur_schema.AttributeType],
    constants: dict[str, float] | None = None,
    logger: logging.Logger | None = None,
    list_return: bool = False
) -> tuple[ConvertedType | list[ConvertedType] | list[ConvertedType | list[ConvertedType]], bool]:
    """
    Tries to converts a given string to the types given in definitions.
    First succeeded conversion will be returned

    If no logger is given and a attribute cannot be converted an error is raised

    :param stringattribute: str, Attribute to convert.
    :param definitions: list of :py:class:`~masci_tools.io.parsers.fleur_schema.AttributeType` definitions
    :param constants: dict, of constants defined in fleur input
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted value of the first successful conversion
    """

    if not isinstance(xmlstring, list):
        xmlstring = [xmlstring]

    if logger is not None:
        logger.debug('Value to convert from XML: %s', xmlstring)

    converted_list: list[ConvertedType | list[ConvertedType]] = []
    all_success = True
    for text in xmlstring:

        split_text = text.split(' ')
        while '' in split_text:
            split_text.remove('')

        text_definitions = []
        for definition in definitions:
            if definition.length == len(split_text):
                text_definitions.append(definition)

        if not text_definitions:
            for definition in definitions:
                if definition.length in ('unbounded', 1):
                    text_definitions.append(definition)

        if not text_definitions:
            if logger is None:
                raise ValueError(f"Could not convert '{text}', no matching definition found")
            logger.warning("Could not convert '%s', no matching definition found", text)
            converted_list.append(text)
            all_success = False
            continue

        types = tuple(definition.base_type for definition in text_definitions)
        lengths = {definition.length for definition in text_definitions}

        if len(text_definitions) == 1:
            if text_definitions[0].length == 1:
                split_text = [text]

        if logger is not None:
            logger.debug('Convert from XML: %s using definitions %s', split_text, types)

        converted_text, suc = convert_from_xml_single_values(split_text, types, constants=constants, logger=logger)

        all_success = all_success and suc

        if len(converted_text) == 1 and 'unbounded' not in lengths:
            converted_text = converted_text[0]  #type:ignore[assignment]
        elif len(converted_text) == 0 and 'unbounded' not in lengths:
            converted_text = ''  #type:ignore

        converted_list.append(converted_text)

    ret_value = converted_list
    if len(converted_list) == 1 and not list_return:
        ret_value = converted_list[0]  #type:ignore[assignment]

    if logger is not None:
        logger.debug('Converted Value from XML: %s', ret_value)

    return ret_value, all_success


def convert_to_xml_explicit(value: Any | Iterable[Any],
                            definitions: list[fleur_schema.AttributeType],
                            logger: logging.Logger | None = None,
                            float_format: str = '.10',
                            list_return: bool = False) -> tuple[str | list[str], bool]:
    """
    Tries to convert a given list of values to str for a xml file based on the definitions (length and type).
    First succeeded conversion will be returned

    :param textvalue: value to convert
    :param definitions: list of :py:class:`~masci_tools.io.parsers.fleur_schema.AttributeType` definitions
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted value of the first successful conversion
    """
    import numpy as np

    lengths = {definition.length for definition in definitions}

    if not isinstance(value, (list, np.ndarray)):
        value = [value]
    elif not isinstance(value[0], (list, np.ndarray)) and lengths != {1}:
        value = [value]

    if logger is not None:
        logger.debug('Value to convert to XML: %s', value)

    converted_list = []
    all_success = True
    for val in value:

        if not isinstance(val, (list, np.ndarray)):
            val = [val]

        text_definitions = []
        for definition in definitions:
            if definition.length == len(val):
                text_definitions.append(definition)

        if not text_definitions:
            for definition in definitions:
                if definition.length == 'unbounded':
                    text_definitions.append(definition)

        if not text_definitions:
            if len(val) == 1 and isinstance(val[0], str):
                converted_list.append(val[0])
                continue

            if logger is None:
                raise ValueError(f"Could not convert '{val}', no matching definition found")
            logger.warning("Could not convert '%s', no matching definition found", val)

            converted_list.append('')
            all_success = False
            continue

        types = tuple(definition.base_type for definition in text_definitions)

        if logger is not None:
            logger.debug('Convert to XML: %s with definitions %s', val, types)

        converted_text, suc = convert_to_xml_single_values(val, types, logger=logger, float_format=float_format)
        all_success = all_success and suc

        converted_list.append(' '.join(converted_text))

    ret_value = converted_list
    if len(converted_list) == 1 and not list_return:
        ret_value = converted_list[0]  #type:ignore

    if logger is not None:
        logger.debug('Converted Value to XML: %s', ret_value)

    return ret_value, all_success


def convert_from_xml_single_values(xmlstring: str | list[str],
                                   possible_types: tuple[BaseType, ...],
                                   constants: dict[str, float] | None = None,
                                   logger: logging.Logger | None = None) -> tuple[list[ConvertedType], bool]:
    """
    Tries to converts a given string attribute to the types given in possible_types.
    First succeeded conversion will be returned

    If no logger is given and a attribute cannot be converted an error is raised

    :param stringattribute: str, Attribute to convert.
    :param possible_types: list of str What types it will try to convert to
    :param constants: dict, of constants defined in fleur input
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted value of the first successful conversion
    """
    from masci_tools.util.fleur_calculate_expression import calculate_expression, MissingConstant

    if not isinstance(xmlstring, list):
        xmlstring = [xmlstring]

    converted_value: ConvertedType
    converted_list = []
    all_success = True
    for text in xmlstring:

        exceptions: list[Exception] = []
        for value_type in possible_types:
            if value_type == 'float':
                try:
                    converted_value = float(text)
                except (ValueError, TypeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'float_expression':
                try:
                    converted_value = calculate_expression(text, constants=constants)
                except ValueError as exc:
                    exceptions.append(exc)
                    continue
                except MissingConstant as exc:
                    new_exc = MissingConstant(f'No value available for expression {exc}\n'
                                              'Please provide the value for this constant'
                                              ' by using the get_constants function for example')

                    exceptions.append(new_exc)
                    continue

            elif value_type == 'complex':
                try:
                    converted_value = convert_from_fortran_complex(text)
                except (ValueError, TypeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'int':
                try:
                    converted_value = int(text)
                except (ValueError, TypeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'switch':
                try:
                    converted_value = convert_from_fortran_bool(text)
                except (ValueError, TypeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'string':
                converted_value = str(text)

            converted_list.append(converted_value)
            break
        else:
            if logger is None:
                raise ValueError(f"Could not convert '{text}'. Tried: {possible_types}.\n"
                                 'The following errors occurred:\n   ' +
                                 '\n   '.join([str(error) for error in exceptions]))
            logger.warning("Could not convert '%s'. The following errors occurred:", text)

            for error in exceptions:
                logger.warning('   %s', str(error))
                logger.debug(error, exc_info=error)

            converted_list.append(text)
            all_success = False

    return converted_list, all_success


def convert_to_xml_single_values(value: Any | Iterable[Any],
                                 possible_types: tuple[BaseType, ...],
                                 logger: logging.Logger | None = None,
                                 float_format: str = '.10') -> tuple[list[str], bool]:
    """
    Tries to converts a given attributevalue to a string for a xml file according
    to the types given in possible_types.
    First succeeded conversion will be returned

    :param value: value to convert.
    :param possible_types: list of str What types it will try to convert from
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted str of the value of the first successful conversion
    """
    import numpy as np

    if not isinstance(value, (list, np.ndarray)):
        value = [value]

    if 'string' not in possible_types:
        possible_types = possible_types + ('string',)  #Always try string

    converted_value: str
    converted_list = []
    exceptions: list[Exception] = []
    all_success = True
    for val in value:

        for value_type in possible_types:
            if value_type in ('float', 'float_expression'):
                if isinstance(val, complex):
                    exceptions.append(ValueError(f'Could not convert {val} to fortran float. Value is complex'))
                    continue
                try:
                    converted_value = f'{val:{float_format}f}'
                except ValueError as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'complex':
                try:
                    converted_value = f'({val.real:{float_format}f},{val.imag:{float_format}f})'
                except (ValueError, AttributeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'switch':
                try:
                    converted_value = convert_to_fortran_bool(val)
                except (ValueError, TypeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type in ('string', 'int'):
                converted_value = str(val)

            converted_list.append(converted_value)
            break
        else:
            if logger is None:
                raise ValueError(f"Could not convert '{val}' to text. Tried: {possible_types}.\n"
                                 'The following errors occurred:\n   ' +
                                 '\n   '.join([str(error) for error in exceptions]))
            logger.warning("Could not convert '%s' to text. The following errors occurred:", val)

            for error in exceptions:
                logger.warning('   %s', str(error))
                logger.debug(error, exc_info=error)

            converted_list.append(val)
            all_success = False

    return converted_list, all_success


def convert_from_fortran_bool(stringbool: str | bool) -> bool:
    """
    Converts a string in this case ('T', 'F', or 't', 'f') to True or False

    :param stringbool: a string ('t', 'f', 'F', 'T')

    :return: boolean  (either True or False)
    """

    true_items = ['True', 't', 'T']
    false_items = ['False', 'f', 'F']
    if isinstance(stringbool, str):
        if stringbool in false_items:
            return False
        if stringbool in true_items:
            return True
        raise ValueError(f"Could not convert: '{stringbool}' to boolean, "
                         "which is not 'True', 'False', 't', 'T', 'F' or 'f'")
    if isinstance(stringbool, bool):
        return stringbool  # no conversion needed...

    raise TypeError(f"Could not convert: '{stringbool}' to boolean, only accepts str or boolean")


def convert_to_fortran_bool(boolean: bool | str) -> Literal['T', 'F']:
    """
    Converts a Boolean as string to the format defined in the input

    :param boolean: either a boolean or a string ('True', 'False', 'F', 'T')

    :return: a string (either 't' or 'f')
    """

    if isinstance(boolean, bool):
        if boolean:
            return 'T'
        return 'F'
    if isinstance(boolean, str):  # basestring):
        if boolean in ('True', 't', 'T'):
            return 'T'
        if boolean in ('False', 'f', 'F'):
            return 'F'
        raise ValueError(f"A string: {boolean} for a boolean was given, which is not 'True',"
                         "'False', 't', 'T', 'F' or 'f'")

    raise TypeError(f'convert_to_fortran_bool accepts only a string or bool as argument, given {boolean} ')


def convert_from_fortran_complex(number_str: str) -> complex:
    """
    Converts a string of the form (float,float) to a complex number

    :param number_str: string to convert

    :returns: complex number
    """
    number_str = number_str.strip()

    RE_COMPLEX_NUMBER = r'\([-+]?(?:\d*\.\d+|\d+)\,[-+]?(?:\d*\.\d+|\d+)\)'
    RE_SINGLE_FLOAT = r'[-+]?(?:\d*\.\d+|\d+)'

    if re.fullmatch(RE_COMPLEX_NUMBER, number_str) is None:
        raise ValueError(f"String '{number_str}' is not of the format (float,float)")
    real_str, imag_str = re.findall(RE_SINGLE_FLOAT, number_str)

    return float(real_str) + 1j * float(imag_str)


def convert_fleur_lo(loelements: list[etree._Element], allow_special_los: bool = True) -> str:
    """
    Converts lo xml elements from the inp.xml file into a lo string for the inpgen
    """
    # Developer hint: Be careful with using '' and "", basestring and str are not the same...
    # therefore other conversion methods might fail, or the wrong format could be written.
    from masci_tools.util.xml.common_functions import get_xml_attribute

    shell_map = {0: 's', 1: 'p', 2: 'd', 3: 'f'}

    lo_string = ''
    for element in loelements:
        lo_type = get_xml_attribute(element, 'type')
        if lo_type != 'SCLO' and not allow_special_los:  # non standard los not supported for now
            continue
        eDeriv = get_xml_attribute(element, 'eDeriv')
        if allow_special_los and eDeriv not in ('0', '1'):  # LOs with higher derivatives are also dropped
            continue
        if not allow_special_los and eDeriv != '0':
            continue
        l_num = get_xml_attribute(element, 'l')
        n_num = get_xml_attribute(element, 'n')
        if l_num is None or n_num is None:
            raise ValueError('Failed to evaluate l and n attribute of LO element')
        lostr = f'{n_num}{shell_map[int(l_num)]}'
        lo_string = lo_string + ' ' + lostr
    return lo_string.strip()


def convert_fleur_electronconfig(econfig_element: etree._Element) -> str:
    """
    Convert electronConfig tag to eConfig string
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.econfig import convert_fleur_config_to_econfig

    core_config = eval_xpath(econfig_element, 'coreConfig/text()')
    valence_config = eval_xpath(econfig_element, 'valenceConfig/text()')

    if not core_config:
        core_config = ''

    if not valence_config:
        valence_config = ''

    core_config_str = convert_fleur_config_to_econfig(cast(str, core_config))
    valence_config_str = convert_fleur_config_to_econfig(cast(str, valence_config))

    return f'{core_config_str} | {valence_config_str}'


def convert_str_version_number(version_str: str) -> tuple[int, int]:
    """
    Convert the version number as a integer for easy comparisons

    :param version_str: str of the version number, e.g. '0.33'

    :returns: tuple of ints representing the version str
    """

    version_numbers = version_str.split('.')

    if len(version_numbers) != 2:
        raise ValueError(f"Version number is malformed: '{version_str}'")

    return tuple(int(part) for part in version_numbers)  #type:ignore



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains Classes for building complex XPath expressions based on
general attribute conditions from simple XPath expressions
"""
from __future__ import annotations

from typing import Any, Iterable
import sys
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from lxml import etree

FilterType: TypeAlias = 'dict[str, Any]'
"""
Type for filters argument for XPathBuilder
"""


class XPathBuilder:
    """
    Class for building a complex xpath (restricted to adding filters)
    from a simple xpath expression

    .. note::
        passing in an etree.XPath object will not respect the options
        passed into it. Only the kwargs in __init__ are used to compile
        the path if compile_path=True

    .. note::
        Filters/Constraints (or predicates like they are called for XPaths) can either
        be added by providing the ``filters`` argument in the constructor or by calling
        the :py:meth:`add_filter()` method.

        The ``filters`` argument is a dictionary with the tag names, where to apply the condition,
        as keys and the condition as values while the :py:meth:`add_filter()` method takes these
        as it's two arguments. The tag name has to be a part of the original simple xpath expression.
        The conditition is a dictionary with one key specifying the kind of condition and the value for the
        condition. The condition can also be the name of an attribute or path, in which case the value can be another
        condition dictionary. The following conditions operators i.e. keys in the dictionary are supported:

            - ``=``/``==``: equal to
            - ``!=``: not equal to
            - ``<``: less than
            - ``>``: greater than
            - ``<=``: less than or equal to
            - ``>=``: greater than or equal to
            - ``contains``: attribute/tag contains the given value (case sensitive)
            - ``not-contains``: attribute/tag does not contains the given value
            - ``starts-with``: attribute/tag starts with the given value (case sensitive)
            - ``ends-with``: attribute/tag ends with the given value (case sensitive)
            - ``index``: Select tags based on their index in the parent tag (either explicit index or another condition)
            - ``has``: Select tags based on the presence of the given attribute/tag
            - ``has-not``: Select tags based on the absence of the given attribute/tag
            - ``number-nodes``: Compute the number of nodes in the previous path and select based on further criteria
            - ``and``: Provide multiple conditions in a list joined by ``and``
            - ``or``: Provide multiple conditions in a list joined by ``or``
            - ``in``: Select tags if the value of the path is in a given list of values
            - ``not-in``: Select tags if the value of the path is not in a given list of values
            - ``<string>``: All other strings are interpreted as paths to attributes/tags specifying conditions on their value
            - ``<tuple of paths>``: Multiple strings are interpreted as multiple node sets, which are joined with |

    Example::

        from masci_tools.util.xml.xpathbuilder import XPathBuilder

        # XPath selecting all lo tags for SCLO type LOs and Iron species
        xpath = XPathBuilder('/fleurInput/atomSpecies/species/lo',
                             filters = {'species': {
                                            'name': {'contains': 'Fe'},
                                        },
                                        'lo': {
                                            'type': 'SCLO'
                                        }
                                    })

    :param simple_path: basic simple XPath expression to start from
    :param filters: dictionary with filters
    :param compile_path: bool if True the path property will be compiled as etree.XPath
    :param strict: bool if True the __str__ conversion will raise an error

    Other Kwargs will be passed on to the etree.XPath compilation if ``compile_path=True``
    """

    BINARY_OPERATORS = {'=', '==', '!=', '<', '>', '<=', '>=', 'contains', 'not-contains', 'starts-with', 'ends-with'}
    UNARY_OPERATORS = {'has', 'has-not', 'number-nodes', 'string-length'}
    COMPOUND_OPERATORS = {'and', 'or', 'in', 'not-in'}

    #Operators, which are not formatted in the default way {left} {operator} {right}
    OPERATOR_FORMAT = {
        '==': '{left} = {right}',
        'contains': 'contains({left}, {right})',
        'not-contains': 'not(contains({left}, {right}))',
        'starts-with': 'starts-with({left}, {right})',
        'ends-with':
        '{right} = substring({left}, string-length({left}) - string-length({right}) + 1)',  #'ends-with({left}, {right})' only available in XPath 2.0
        'number-nodes': 'count({left})',
        'string-length': 'string-length({left})',
        'has': '{left}',
        'has-not': 'not({left})',
    }

    def __init__(self,
                 simple_path: str | bytes | etree.XPath,
                 filters: dict[str, FilterType] | None = None,
                 compile_path: bool = False,
                 strict: bool = False,
                 **kwargs: Any) -> None:
        self.compile_path = compile_path
        self.strict = strict
        if not self.compile_path and kwargs:
            raise ValueError('Keyword arguments only available for compiled Xpaths')
        if isinstance(simple_path, str):
            self.components = simple_path.rstrip('/').split('/')
        elif isinstance(simple_path, etree.XPath):
            self.components = simple_path.path.split('/')
        else:
            raise TypeError(f'Wrong type for simple path. Expected str or etree.XPath. Got {type(simple_path)}')

        if len(set(self.components)) != len(self.components):
            raise NotImplementedError('The given xpath has multiple tags with the same name')

        self.path_kwargs = kwargs
        self.filters: dict[str, FilterType] = {}
        self.path_variables: dict[str, etree._XPathObject] = {}
        self.value_conditions: int = 0
        if filters is not None:
            for key, val in filters.items():
                self.add_filter(key, val)

    def add_filter(self, tag: str, conditions: FilterType | Any) -> None:
        """
        Add a filter to the filters dictionary

        :param tag: str name of the tag name to add a filter to
        :param conditions: dictionary specifying the filter
        """
        if tag not in self.components:
            raise ValueError(f"The tag {tag} is not part of the given xpath expression: {'/'.join(self.components)}")
        if not isinstance(conditions, dict):
            conditions = {'=': conditions}

        self.filters[tag] = {**self.filters.get(tag, {}), **conditions}

    def append_tag(self, tag: str) -> None:
        """
        Append another tag to the end of the simple xpath expression

        :param tag: str name of the tag to append
        """
        if tag in self.components:
            raise NotImplementedError(
                f"The tag {tag} is already part of the given xpath expression: {'/'.join(self.components)}")

        self.components.append(tag)

    def strip_off_tag(self) -> str:
        """
        Strip off the last tag of the simple xpath expression
        """
        if not self.components:
            raise ValueError('Cannot strip off tag. Path is empty')

        #TODO: Check if the filters contains filters for the last tag

        return self.components.pop(-1)

    def get_predicate(self,
                      tag: str,
                      condition: Any,
                      compound: bool = False,
                      path: str | tuple[str, ...] = '.',
                      process_path: bool = False) -> str:
        """
        Construct the predicate for the given tag and condition

        :param tag: str name of the tag
        :param condition: condition specified, either dict or single value
        :param compound: bool if True the enclosing condition is a compound condition, forbidding any other compound condition
        :param path: path, to which to apply the condition
        :param process_path: bool if True the path will taken apart into its components and the components will be checked with XPath variables
        """

        if not isinstance(condition, dict):
            condition = {'=': condition}

        if len(condition) > 1:
            raise ValueError('Only one key allowed in condition')

        operator, content = dict(condition).popitem()

        if compound and operator in self.COMPOUND_OPERATORS:
            raise ValueError(f'Compound operators not allowed in already compound condition: {operator}')

        return self.process_condition(tag, operator, content, path, process_path=process_path)

    def process_condition(self,
                          tag: str,
                          operator: str,
                          content: Any,
                          path: str | tuple[str, ...],
                          process_path: bool = False) -> str:
        """
        Process the condition for the given tag and condition

        :param tag: str name of the tag
        :param operator: operator for condition
        :param content: content of condition
        :param path: path, to which to apply the condition
        :param process_path: bool if True the path will taken apart into its components and the components will be checked with XPath variables
        """

        if operator == 'and':
            if not isinstance(content, (list, tuple)):
                raise TypeError('For and operator provide the conditions as a list')
            predicates = [
                self.get_predicate(tag, condition_part, compound=True, path=path, process_path=process_path)
                for condition_part in content
            ]
            predicate = ' and '.join(predicates)
        elif operator == 'or':
            if not isinstance(content, (list, tuple)):
                raise TypeError('For or operator provide the conditions as a list')
            predicates = [
                self.get_predicate(tag, condition_part, compound=True, path=path, process_path=process_path)
                for condition_part in content
            ]
            predicate = ' or '.join(predicates)
        elif operator == 'in':
            if not isinstance(content, Iterable):
                raise TypeError('For in operator provide a sequence of possible values')
            predicate = self.get_predicate(tag, {'or': [{
                '=': value
            } for value in content]},
                                           path=path,
                                           process_path=process_path)
        elif operator == 'not-in':
            if not isinstance(content, Iterable):
                raise TypeError('For not-in operator provide a sequence of possible values')
            predicate = self.get_predicate(tag, {'and': [{
                '!=': value
            } for value in content]},
                                           path=path,
                                           process_path=process_path)
        elif operator == 'index':
            if isinstance(content, int):
                index = content
                if index == -1:
                    predicate = 'last()'
                elif index < 0:
                    index = abs(index + 1)
                    predicate = f'last() - ${tag}_index'
                else:
                    predicate = f'${tag}_index'
            else:
                cond, index = dict(content).popitem()
                if cond not in self.BINARY_OPERATORS:
                    raise ValueError(f'Operator {cond} not allowed for index')
                if index < 0:
                    index = abs(index + 1)
                    index_str = f'last() - ${tag}_index'
                else:
                    index_str = f'${tag}_index'
                operator_fmt = self.OPERATOR_FORMAT.get(cond, f'{{left}} {cond} {{right}}')
                predicate = operator_fmt.format(left='position()', right=index_str)
            self.path_variables[f'{tag}_index'] = index
        elif operator in self.BINARY_OPERATORS:
            operator_fmt = self.OPERATOR_FORMAT.get(operator, f'{{left}} {operator} {{right}}')

            variable_name = f'{tag}_cond_{self.value_conditions}'
            if process_path:
                path_variable = self._path_condition(path, tag)
            else:
                path_variable = ' | '.join(path) if isinstance(path, tuple) else path

            predicate = operator_fmt.format(left=path_variable, right=f'${variable_name}')
            self.path_variables[variable_name] = content
            self.value_conditions += 1
        elif operator in self.UNARY_OPERATORS - {'has', 'has-not'}:
            operator_fmt = self.OPERATOR_FORMAT.get(operator, '{left}')
            if process_path:
                path_variable = self._path_condition(path, tag)
            else:
                path_variable = ' | '.join(path) if isinstance(path, tuple) else path
            path_variable = operator_fmt.format(left=path_variable)
            predicate = self.get_predicate(tag, content, path=path_variable)
        elif operator in {'has', 'has-not'}:
            operator_fmt = self.OPERATOR_FORMAT.get(operator, '{left}')
            predicate = operator_fmt.format(left=self._path_condition(content, tag))
        else:
            predicate = self.get_predicate(tag, content, path=operator, process_path=True)

        return predicate

    def _path_condition(self, path: str | tuple[str, ...], prefix: str) -> str:
        """
        Prepare conditions based on variables

        :param path: str path to process
        :param prefix: str prefix to use for xpath variables
        """
        if isinstance(path, tuple):
            #Multiple node sets. Join them with union
            return ' | '.join(self._path_condition(p, f'{prefix}_{indx}') for indx, p in enumerate(path))

        parts = path.strip('/').split('/')
        variable_name = f'{prefix}_cond_{self.value_conditions}_name'

        path_variable = []
        for indx, part in enumerate(parts):
            if part != '.':
                path_variable.append(
                    f"{'@' if '@' in part or '/' not in path else ''}*[local-name()=${variable_name}_{indx}]")
            else:
                path_variable.append(part)
            self.path_variables[f'{variable_name}_{indx}'] = part.lstrip('@')

        return '/'.join(path_variable)

    @property
    def path(self) -> str | etree.XPath:
        """
        Property for constructing the complex Xpath
        """

        predicates = [''] * len(self.components)
        self.path_variables = {}
        self.value_conditions = 0

        for tag, condition in self.filters.items():
            component_index = self.components.index(tag)
            predicates[component_index] = self.get_predicate(tag, condition)

        path = '/'.join(
            [f'{tag}[{predicate}]' if predicate else tag for tag, predicate in zip(self.components, predicates)])
        if self.compile_path:
            return etree.XPath(path, **self.path_kwargs)
        return path

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({'/'.join(self.components)!r}, {self.filters!r}, compile_path={self.compile_path!r}, strict={self.strict!r})"

    def __str__(self) -> str:
        if self.strict:
            raise ValueError(f'Implicit string conversion for {self.__class__.__qualname__}.')
        path = self.path
        if isinstance(path, etree.XPath):
            path = path.path

        for name, value in sorted(self.path_variables.items(), key=lambda x: len(x[0]), reverse=True):
            path = path.replace(f'${name}', f'{value!r}')
        return path



# pylint: disable=unused-import
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""This subpackage contains recipes for the tabulator subpackage, which turns
properties of a collections of objects into a table.

Recipes let you reuse tabulator settings for different use cases.
"""

import abc as _abc
import typing as _typing

import masci_tools.util.python_util as _masci_python_util
from .transformers import Transformer


class Recipe(_abc.ABC):
    """Recipe for a :py:class:`~.tabulator.Tabulator`.

    Recipes hold the include, exclude list of properties which a tabulator should put into a table, by reading
    them from a set of objects, one row per object. In other words, the recipe specifies the column of the table.
    Transformations of properties for the table (say, a property is a list, and we only want the maximum), can
    be defined by specifying a transformer in the recipe.

    TODO: improve memory performance for tabulation:

    - let include exclude lists values optionally hold pandas dtype strings (numpy 'uint8' etc plus pandas types like
      'categorical'; the latter internally in tabulator replaced by numpy correspondence where needed). tabulator
      will use those when building table. otherwise, e.g. when returning pandas dataframe, all columns will
      have dtype 'object' or 'float64' and the table won't fit into memory anymore very quickly.
    """

    def __init__(self, exclude_list: dict = None, include_list: dict = None, transformer: Transformer = None, **kwargs):
        """Initialize a recipe for a :py:class:`~.tabulator.Tabulator`.

        The attributes :py:attr:`~.include_list` and :py:attr:`~.exclude_list` control which properties
        are to be tabulated (=table columns). A tabulator defines an `autolist` method which can auto-generate
        an extensive include list from a given object. This can be used to define a custom include or exclude list
        for objects of that use case.

        Subclasses define the nature of the objects to be tabulated by making assumptions on their
        property structure. That way, if both include and exclude list are empty, by default the 'complete'
        set of properties of the objects will be tabulated, where the subclass defines the notion of 'complete'.

        If neither exclude nor include list is given, the full set of properties according to implementation
        will be tabulated.

        There are three accepted formats for the include and exclude lists, here shown for an example of an
        AiiDA workchain node (for aiida-jutools `NodeTabulator`). In this example, the objects to be a tabulated
        are nodes in a database, and they have a unique identifier ('uuid'), a label, and some attributes called
        'extras' and 'outputs' with many nested properties inside them. The selection here will tell a tabulator
        to only put those properties in a table.

        .. code-block:: python

           # format 1: 'none' format: all keys are mentioned, values are None
           include_withNone = {
               'uuid': None,
               'extras':{
                   'scale_factor':None,
                   'kkr_constants_version' : {
                       'constants_version': None
                   }
               },
               "outputs": {
                   "last_calc_info": {
                       "convergence_reached": None,
                   },
                   "last_calc_output_parameters": {
                       "charge_core_states_per_atom": None,
                       "charge_core_states_per_atom_unit": None,
                       "charge_valence_states_per_atom": None,
                       "charge_valence_states_per_atom_unit": None,
                   }
               }
           }

           # format 2: 'list' format: if all subkeys remain on the same level, can
           # use shorthand notation and write them as a list instead of
           # {key:None, key:None, ...}
           include_withList = {
               'uuid': None,
               'extras':{
                   'scale_factor': None,
                   'kkr_constants_version' : ['constants_version']
               },
               "outputs": {
                   "last_calc_info": ["convergence_reached"],
                   "last_calc_output_parameters": [
                       "charge_core_states_per_atom",
                       "charge_core_states_per_atom_unit",
                       "charge_valence_states_per_atom",
                       "charge_valence_states_per_atom_unit"
                   ]
               }

           }

           # format 3: keypaths. this is the internal format the recipe uses.
           include_keypaths = [
               ['uuid'],
               ['extras', 'scale_factor'],
               ['extras', 'kkr_constants_version', 'constants_version'],
               ['outputs', 'last_calc_info', 'convergence_reached'],
               ['outputs', 'last_calc_output_parameters', 'charge_core_states_per_atom'],
               ['outputs', 'last_calc_output_parameters', 'charge_core_states_per_atom_unit'],
               ['outputs', 'last_calc_output_parameters', 'charge_valence_states_per_atom'],
               ['outputs', 'last_calc_output_parameters', 'charge_valence_states_per_atom_unit']
           ]

        :param exclude_list: Optional list of properties to exclude. May be set later.
        :param include_list: Optional list of properties to include. May be set later.
        :param transform: Specifies special transformations for certain properties for tabulation.
        :param kwargs: Additional keyword arguments for subclasses.
        """
        # note: for the in/ex lists, using the public setter here,
        # to trigger conversion
        self._exclude_list = exclude_list if exclude_list else {}
        self._include_list = include_list if include_list else {}
        self.transformer = transformer

    @property
    def exclude_list(self) -> dict:
        return self._exclude_list

    @exclude_list.setter
    def exclude_list(self, exclude_list: _typing.Union[dict, list]):
        self._exclude_list = exclude_list
        if isinstance(exclude_list, dict):
            self._to_keypaths()

    @property
    def include_list(self) -> dict:
        return self._include_list

    @include_list.setter
    def include_list(self, include_list: _typing.Union[dict, list]):
        self._include_list = include_list
        if isinstance(include_list, dict):
            self._to_keypaths()

    def _to_keypaths(self):
        """Generate paths from a possibly nested dictionary.

        This method can be used for handling include lists, exclude lists, and when writing
        new :py:class:`~Transformer` transform methods.

        List of paths to each value within the dict as tuples (path, value).

        convert from with-List to with-None format for convert to keypaths

        convert to keypaths (upper: done inside this one anyway)
        """

        def _to_keypaths_recursive(sub_dict: dict, path: list):
            paths = []
            for k, v in sub_dict.items():
                if isinstance(v, dict):
                    paths += _to_keypaths_recursive(v, path + [k])
                paths.append((path + [k], v))
            return paths

        for in_or_ex, a_dict in {'in': self._include_list, 'out': self._exclude_list}.items():

            # precondition: not already keypaths format
            is_list = isinstance(a_dict, list)
            is_all_lists = is_list and all(isinstance(path, list) for path in a_dict)
            if is_all_lists:
                continue

            # if empty, convert to empty list. if not empty, convert to keypaths
            if not a_dict:
                keypaths = []
            else:
                # convert from include list with-list format with-none format:
                # same-level subkeys mentioned as list [k1,k2] -> dict {k1:None, k2:None}.
                _a_dict = _masci_python_util.modify_dict(a_dict=a_dict,
                                                         transform_value=lambda v: {k: None for k in v}
                                                         if isinstance(v, list) else v,
                                                         to_level=99)

                keypaths = _to_keypaths_recursive(sub_dict=_a_dict, path=[])
                # the result consists of sets of subpaths. For each subset, there is
                # an additianal entry where the value contains the whole subdict from
                # which the paths were generated. We are not interested in those duplicate
                # entries, so remove them.
                keypaths = [tup for tup in keypaths if not isinstance(tup[1], dict)]

                # now list should be like [(path1, None), (path2, None), ...],
                # or at least of type _typing.List[_typing.Tuple[list, _typing.Any]].
                # check that. if not, something is wrong.
                # otherwise, just return the paths.
                if all(tup[1] is None for tup in keypaths):
                    keypaths = [tup[0] for tup in keypaths]

            # postcondition: keypaths format
            is_list = isinstance(keypaths, list)
            is_all_lists = is_list and all(isinstance(path, list) for path in keypaths)
            if not is_all_lists:
                raise TypeError(f'Could not generate keypaths of required type {_typing.List[list]} '
                                f'from {in_or_ex}clude list. Either specified list in wrong format '
                                f'(see class init docstring for examples), or list generated from '
                                f'autolist stumbled over untreated special case for some unpacked '
                                f'property.')

            if in_or_ex == 'in':
                self._include_list = keypaths
            elif in_or_ex == 'out':
                self._exclude_list = keypaths



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""Convenience tools only related to Python Standard Library."""

from __future__ import annotations

import copy as _copy
import dataclasses as _dc
import datetime as _datetime
import json as _json
import random as _random
import re as _re
import string as _string
import sys as _sys
import typing as _typing

import humanfriendly as _humanfriendly
import pytz as _pytz


def is_number(a_string: str) -> bool:
    """Checks if string represents an int or float. Supports leading/trailing whitespaces, scientific notation.

    Reference: https://stackoverflow.com/a/23639915/8116031
    :param a_string: a string
    :return: True if string represents int or float, False otherwise.
    """
    # DEVNOTE: the same kind of test can be used for numpy numerical types.
    try:
        float(a_string)
        return True
    except ValueError:
        return False


def to_number(a_string: str) -> int | float | str:
    """Converts a string representation of a number into a numerical type.

    Numbers with decimals below machine epsilon, usually around 1e-16, get recognized as int.

    :param a_string: a string
    :return: Float if represents float, int if represents int, string otherwise.
    :rtype: float, int or str.
    """
    num: str | float | int = a_string
    try:
        num = float(num)
        num = int(num) if num.is_integer() else num
    except ValueError:
        pass
    return num


def now() -> _datetime.datetime:
    """Get now time localized to UTC.

    This is the same format which AiiDA uses for attributes `ctime`, `mtime`.
    """
    return _datetime.datetime.now(tz=_pytz.UTC)
    # same as: pytz.UTC.localize(datetime.now())
    # NOT same as: datetime.now(tz=timezone.utc)
    # NOT same as: datetime.utcnow()


def validate_datetime_isoformat(datetime_str: str) -> bool:
    """Validate whether a datetime object string is isoformat.

    :param datetime_str: A datetime string.
    :return: True if isoformat, else False.

    >>> import datetime
    >>> from masci_tools.util.python_util import validate_datetime_isoformat
    >>>
    >>> assert validate_datetime_isoformat(datetime.datetime.now().isoformat())
    >>> assert validate_datetime_isoformat(datetime.datetime.now().isoformat('T','seconds'))
    """
    regex = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$'
    match = _re.compile(regex).match
    return match(datetime_str) is not None


def random_string(length: int,
                  ascii_uppercase: bool = True,
                  ascii_lowercase: bool = True,
                  digits: bool = True,
                  whitespace: bool = False) -> str:
    """Generate a random string of length.

    :param length: random string length.
    :param ascii_uppercase: True: include ASCII uppercase letters.
    :param ascii_lowercase: True: include ASCII lowercase letters.
    :param digits: True: include digits.
    :param whitespace: True: include whitespaces `\t\n\r\x0b\x0c`.
    :return: random string of length.
    """
    character_sets = {
        _string.ascii_uppercase: ascii_uppercase,
        _string.ascii_lowercase: ascii_lowercase,
        _string.digits: digits,
        _string.whitespace: whitespace
    }
    population = ''.join(characters for characters, included in character_sets.items() if included)

    return ''.join(_random.choices(population=population, k=length))


def enforce_minimum_python_version(minimum_version: tuple[int, int] = (3, 7)) -> int:
    """For some operation requiring a minimum version, enforce for interpreter.

    :param minimum_version:
    :return: 0 if satisfied, sys.exit if not
    """
    current_version = _sys.version_info
    if (current_version[0] != minimum_version[0] or current_version[1] < minimum_version[1]):
        _sys.stderr.write('[%s] - Error: Your Python interpreter must be %d.%d or greater (within major version %d)\n' %
                          (_sys.argv[0], minimum_version[0], minimum_version[1], minimum_version[0]))
        _sys.exit(-1)
    return 0


def flatten_list_of_lists(list_of_lists: list[list[_typing.Any]]) -> list[_typing.Any]:
    return [item for sublist in list_of_lists for item in sublist]


def split_list(lst: list, n: int) -> _typing.Generator[_typing.Any, None, None]:
    """Yield n-sized chunks/batches from lst. Last chunk has remainder size.

    :param lst: list
    :param n: chunk size
    :return: split list (copy)
    :rtype: generator

    DEVNOTES:
    - equivalent list comprehension: [lst[i:i + n] for i in range(0, len(lst), n)]
    - as soon as a function contains 'yield', it always returns a generator. so cannot add a switch 'as_list'.
      But can just wrap the call inside list().
    """
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def sort_lists_by_list(lists: list[list], key_list_index: int = 0, desc: bool = False) -> _typing.Iterator[tuple]:
    """Sort multiple lists by one list.

    The key_list gets sorted. The other lists get sorted in the same order.

    Reference: https://stackoverflow.com/a/15611016/8116031

    :param lists: list of lists to be sorted
    :param key_list_index: index of list in lists to sort by
    :param desc: False: ascending order
    :return: sorted lists
    """
    # note: two-liner equivalent:
    # line1: convert (zip(a,b,c,...)) lists into list of tuples, and
    #        sort elements of this list by key_list elements, i.e. key_list-th
    #        element in each tuple.
    # line2: unpack tuple list back into original, sorted, lists using
    #        nested list comprehensions and tuple unpacking.
    # sorted_lists = sorted(zip(*lists), reverse=desc, key=lambda x: x[key_list])
    # return [[x[i] for x in sorted_lists] for i in range(len(lists))]

    return zip(*sorted(zip(*lists), reverse=desc, key=lambda x: x[key_list_index]))


def sort_dict(a_dict: dict, key: _typing.Callable | None = None, reverse: bool = False) -> dict:
    """Return a copy of a dictionary, sorted by its keys. Nested / recursive.

    :param a_dict: dictionary
    :param key: optional function to use for sorting, default `None`. Example: `len` for string keys.
    :param reverse: False: sort ascending, True: descending.
    :return: sorted copy of dictionary
    """
    return {k: sort_dict(v) if isinstance(v, dict) else v for k, v in sorted(a_dict.items(), key=key, reverse=reverse)}


def modify_dict(a_dict: dict, transform_value: _typing.Callable = lambda v: v, to_level: int = 99) -> dict:
    """Return a copy of a dictionary with modified values. Nested / recursive.

    Example:

    >>> from masci_tools.util.python_util import modify_dict, sort_dict
    >>>
    >>> d = {'b': [2, 3],
    >>>      'a': 1,
    >>>      'c': {'ca': 4, 'cb': [5, 6], 'cc': {'ccb': [8, 9], 'cca': 7}}}
    >>>
    >>> d2 = {'a': 1,
    >>>       'b': {2: None, 3: None},
    >>>       'c': {'ca': 4, 'cb': {5: None, 6: None}, 'cc': {'cca': 7, 'ccb': [8, 9]}}}
    >>>
    >>> d3 = sort_dict(modify_dict(a_dict=d,
    >>>                            transform_value=lambda v: {k:None for k in v} if isinstance(v,list) else v,
    >>>                            to_level=2)
    >>>               )
    >>> d3 == d2
    True

    :param a_dict: dictionary
    :param transform_value: function to transform (each non-dict) value. Default returns value unchanged.
    :param to_level: Stop modifications below this level. The values on uppermost level have level 1.
    :return: modified copy of dictionary
    """
    to_level = to_level if to_level else int(1e9)

    def inner_modify_dict(sub_dict: dict, level: int = 1) -> dict:
        if level <= to_level:
            for k, v in sub_dict.copy().items():
                if isinstance(v, dict):
                    sub_dict[k] = inner_modify_dict(v, level + 1)
                else:
                    sub_dict[k] = transform_value(v)
        return sub_dict

    copy_dict = _copy.deepcopy(a_dict)
    return inner_modify_dict(copy_dict)


class NoIndent:
    """ Value wrapper. """

    def __init__(self, value: _typing.Any):
        self.value = value


class JSONEncoderTailoredIndent(_json.JSONEncoder):
    """JSONEncoder which allows to not indent items wrapped in :py:class:`~.NoIndent`.

    Reference: https://stackoverflow.com/a/13252112/8116031

    :param kwargs: These are passed on to the json.JSONENcoder

    >>> from masci_tools.util.python_util import JSONEncoderTailoredIndent, NoIndent
    >>> import json
    >>> from string import ascii_lowercase as letters
    >>> data_structure = {
    >>>     'layer1': {
    >>>         'layer2': {
    >>>             'layer3_1': NoIndent([{"x":1,"y":7}, {"x":0,"y":4}, {"x":5,"y":3},
    >>>                                   {"x":6,"y":9},
    >>>                                   {k: v for v, k in enumerate(letters)}]),
    >>>             'layer3_2': 'string',
    >>>             'layer3_3': NoIndent([{"x":2,"y":8,"z":3}, {"x":1,"y":5,"z":4},
    >>>                                   {"x":6,"y":9,"z":8}]),
    >>>             'layer3_4': NoIndent(list(range(20))),
    >>>         }
    >>>     }
    >>> }
    >>> print(json.dumps(data_structure, cls=JSONEncoderTailoredIndent, sort_keys=True, indent=2))
    """

    FORMAT_SPEC = '@@{}@@'
    regex = _re.compile(FORMAT_SPEC.format(r'(\d+)'))

    def __init__(self, **kwargs):
        # Save copy of any keyword argument values needed for use here.
        self.__sort_keys = kwargs.get('sort_keys', None)
        super().__init__(**kwargs)

    def default(self, o: _typing.Any) -> _typing.Any:  # pylint: disable=arguments-differ
        if isinstance(o, NoIndent):
            return self.FORMAT_SPEC.format(id(o))
        else:
            return super().default(o)

    def encode(self, o: _typing.Any) -> str:  # pylint: disable=arguments-differ
        from _ctypes import PyObj_FromPtr  #type: ignore[import]

        format_spec = self.FORMAT_SPEC  # Local var to expedite access.
        json_repr = super().encode(o)  # Default JSON.

        # Replace any marked-up object ids in the JSON repr with the
        # value returned from the json.dumps() of the corresponding
        # wrapped Python object.
        for match in self.regex.finditer(json_repr):
            # see https://stackoverflow.com/a/15012814/355230
            id_obj = int(match.group(1))
            no_indent = PyObj_FromPtr(id_obj)
            json_obj_repr = _json.dumps(no_indent.value, sort_keys=self.__sort_keys)

            # Replace the matched id string with json formatted representation
            # of the corresponding Python object.
            json_repr = json_repr.replace(f'"{format_spec.format(id_obj)}"', json_obj_repr)

        return json_repr


class JSONEncoderDatetime2Isoformat(_json.JSONEncoder):
    """JSONEncoder which serializes datetime values into dateime isoformat strings.

    To deserialize back from isoformat string to datetime object, write a method which
    when meeting such a value, turns it into a datetime object, and pass it to the used JSON
    load function as `object_hook` argument.

    >>> import datetime
    >>> import json
    >>> from masci_tools.util.python_util import JSONEncoderDatetime2Isoformat
    >>>
    >>> # define decoder isoformat -> datetime
    >>> def json_decode_isoformat_2_datetime(a_dict):
    >>>     date_keys = ['ctime', 'mtime']
    >>>     if any(key in a_dict for key in date_keys):
    >>>         for key in date_keys:
    >>>             value = a_dict.get(key, None)
    >>>             if value:
    >>>                 a_dict[key] = _datetime.datetime.fromisoformat(value)
    >>>     return a_dict
    >>>
    >>> # de/serialization roundtrip
    >>> foo = {'ctime': now(), 'bar': 42, 'mtime': datetime.datetime(2021, 9, 28)}
    >>> foo_enc = json.dumps(foo, indent=4, cls=JSONEncoderDatetime2Isoformat)
    >>> foo_dec = json.loads(foo_enc, object_hook=json_decode_isoformat_2_datetime)
    >>> assert foo_dec == foo
    """

    def default(self, o: _typing.Any) -> _typing.Any:
        if isinstance(o, (_datetime.date, _datetime.datetime)):
            return o.isoformat()


_T = _typing.TypeVar('_T')
"""
Generic type alias
"""


def dataclass_default_field(obj: _T, deepcopy: bool = True) -> _T:
    """Abbreviator for defining mutable types as default values for fields of dataclasses.
    References:
    - https://stackoverflow.com/a/53870411/8116031
    - https://docs.python.org/3/library/dataclasses.html#mutable-default-values

    Use case for this: the default value is some class instance. If you just write,
    in dataclass X, y : Y = dc.field(default=Y()), all x, x2, ... instances of X's
    variables x.y will be the same object / instance of Y. Whereas with this method, all
    x.y, x2.y, ... will be different objects / instances of Y (if deepcopy). Apart from that,
    this method allows you to also define a non-default instance of Y (ie, non-empty constructor)
    like Y(foo='bar').

    :param obj: complex / mutable default value
    :param deepcopy: default True: perform deepcopy, False: perform shallow copy.
    :return: dataclass field default value

    >>> import dataclasses as dc
    >>> from typing import Dict, Tuple
    >>> @dc.dataclass
    >>> class C:
    >>>     complex_attribute: Dict[str, Tuple[int, str]] = \
    >>>     dataclass_default_field({"a": (1, "x"), "b": (1, "y")})
    """
    if deepcopy:
        return _dc.field(default_factory=lambda: _copy.deepcopy(obj))
    return _dc.field(default_factory=lambda: _copy.copy(obj))


class SizeEstimator:
    """Container for various python in-memory object size estimation methods.

    Estimating Python object size is not straightforward.

    There are many solutions available online. This is just a small referenced collection of them.
    """

    def __init__(self):
        self.sys = _sys
        self.hf = _humanfriendly

        # Custom objects know their class.
        # Function objects seem to know way too much, including modules.
        # Exclude modules as well.
        from types import ModuleType, FunctionType
        self.BLACKLIST = type, ModuleType, FunctionType

    def sizeof_simple(self, obj: _typing.Any) -> tuple[int, str]:
        """For simple PSL data structures. Reference: https://stackoverflow.com/a/19865746/8116031

        :return: tuple of bytesize, humanfriendly bytesize
        """
        size = self.sys.getsizeof(obj)
        return size, self.hf.format_size(size)

    def sizeof_via_blacklist(self, obj: _typing.Any) -> tuple[int, str]:
        """sum size of object & members. Reference: https://stackoverflow.com/a/30316760/8116031."""
        from gc import get_referents

        if isinstance(obj, self.BLACKLIST):
            raise TypeError('getsize() does not take argument of type: ' + str(type(obj)))
        seen_ids = set()
        size = 0
        objects = [obj]
        while objects:
            need_referents = []
            for obj in objects:  # pylint: disable=redefined-argument-from-local
                if not isinstance(obj, self.BLACKLIST) and id(obj) not in seen_ids:
                    seen_ids.add(id(obj))
                    size += self.sys.getsizeof(obj)
                    need_referents.append(obj)
            objects = get_referents(*need_referents)

        return size, self.hf.format_size(size)

    def sizeof_via_whitelist(self, obj: _typing.Any) -> tuple[int, str]:
        """Recursively iterate to sum size of object & members.

        'gives much more fine-grained control over the types we're going to count for memory usage,
        but has the danger of leaving types out'

        Reference: https://stackoverflow.com/a/30316760/8116031.
        """
        from numbers import Number
        from collections import deque  # pylint: disable=no-name-in-module
        from collections.abc import Set, Mapping

        zero_depth_bases = (str, bytes, Number, range, bytearray)
        _seen_ids = set()

        def inner(obj):
            obj_id = id(obj)
            if obj_id in _seen_ids:
                return 0
            _seen_ids.add(obj_id)
            size = self.sys.getsizeof(obj)
            if isinstance(obj, zero_depth_bases):
                pass  # bypass remaining control flow and return
            elif isinstance(obj, (tuple, list, Set, deque)):
                size += sum(inner(i) for i in obj)
            elif isinstance(obj, Mapping) or hasattr(obj, 'items'):
                size += sum(inner(k) + inner(v) for k, v in getattr(obj, 'items')())
            # Check for custom object instances - may subclass above too
            if hasattr(obj, '__dict__'):
                size += inner(vars(obj))
            if hasattr(obj, '__slots__'):  # can have __slots__ with __dict__
                size += sum(inner(getattr(obj, s)) for s in obj.__slots__ if hasattr(obj, s))
            return size

        size = inner(obj)
        return size, self.hf.format_size(size)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Here are general and special bokeh plots to use

"""
from .bokeh_plotter import BokehPlotter
from .parameters import ensure_plotter_consistency, NestedPlotParameters
from .data import process_data_arguments

import pandas as pd
import numpy as np
import warnings
from pprint import pprint

################## Helpers     ################

plot_params = BokehPlotter()


def set_bokeh_plot_defaults(**kwargs):
    """
    Set defaults for bokeh backend
    according to the given keyword arguments

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    plot_params.set_defaults(**kwargs)


def reset_bokeh_plot_defaults():
    """
    Reset the defaults for bokeh backend
    to the hardcoded defaults

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    plot_params.reset_defaults()


def show_bokeh_plot_defaults():
    """
    Show the currently set defaults for bokeh backend

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    pprint(plot_params.get_dict())


def get_bokeh_help(key):
    """
    Print the description of the given key in the bokeh backend

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    plot_params.get_description(key)


def load_bokeh_defaults(filename='plot_bokeh_defaults.json'):
    """
    Load defaults for the bokeh backend from a json file.

    :param filename: filename,from  where the defaults should be taken
    """
    plot_params.load_defaults(filename)


def save_bokeh_defaults(filename='plot_bokeh_defaults.json', save_complete=False):
    """
    Save the current defaults for the matplotlib backend to a json file.

    :param filename: filename, where the defaults should be stored
    :param save_complete: bool if True not only the overwritten user defaults
                          but also the unmodified hardcoded defaults are stored
    """
    plot_params.save_defaults(filename, save_complete=save_complete)


##################################### general plots ##########################


@ensure_plotter_consistency(plot_params)
def bokeh_scatter(x,
                  y=None,
                  *,
                  xlabel='x',
                  ylabel='y',
                  title='',
                  figure=None,
                  data=None,
                  saveas='scatter',
                  copy_data=False,
                  **kwargs):
    """
    Create an interactive scatter plot with bokeh

    :param x: arraylike or key for data for the x-axis
    :param y: arraylike or key for data for the y-axis
    :param data: source for the data of the plot (pandas Dataframe for example)
    :param xlabel: label for the x-axis
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure
    :param outfilename: filename of the output file
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:class:`masci_tools.vis.bokeh_plotter.BokehPlotter`.
    If the arguments are not recognized they are passed on to the bokeh function `scatter`
    """
    from bokeh.models import ColumnDataSource

    if isinstance(x, (dict, pd.DataFrame, ColumnDataSource)) or x is None:
        warnings.warn(
            'Passing the source as first argument is deprecated. Please pass in source by the keyword data'
            'and xdata and ydata as the first arguments', DeprecationWarning)
        data = x
        x = kwargs.pop('xdata', 'x')
        y = kwargs.pop('ydata', 'y')

    plot_data = process_data_arguments(data=data,
                                       x=x,
                                       y=y,
                                       copy_data=copy_data,
                                       single_plot=True,
                                       same_length=True,
                                       use_column_source=True)
    entry, source = plot_data.items(first=True)

    plot_params.set_defaults(default_type='function', name=entry.y)
    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)

    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    plot_kwargs = plot_params.plot_kwargs(plot_type='scatter')
    res = p.scatter(x=entry.x, y=entry.y, source=source, **plot_kwargs, **kwargs)
    plot_params.add_tooltips(p, res, entry)

    if plot_params['level'] is not None:
        res.level = plot_params['level']

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.save_plot(p, saveas)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_multi_scatter(x,
                        y=None,
                        *,
                        data=None,
                        figure=None,
                        xlabel='x',
                        ylabel='y',
                        title='',
                        saveas='scatter',
                        copy_data=False,
                        set_default_legend=True,
                        **kwargs):
    """
    Create an interactive scatter (muliple data sets possible) plot with bokeh

    :param x: arraylike or key for data for the x-axis
    :param y: arraylike or key for data for the y-axis
    :param data: source for the data of the plot (pandas Dataframe for example)
    :param xlabel: label for the x-axis
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure
    :param outfilename: filename of the output file
    :param copy_data: bool, if True the data argument will be copied
    :param set_default_legend: bool if True the data names are used to generate default legend labels

    Kwargs will be passed on to :py:class:`masci_tools.vis.bokeh_plotter.BokehPlotter`.
    If the arguments are not recognized they are passed on to the bokeh function `scatter`
    """
    from bokeh.models import ColumnDataSource

    if isinstance(x, (dict, pd.DataFrame, ColumnDataSource)) or x is None:
        warnings.warn(
            'Passing the source as first argument is deprecated. Please pass in source by the keyword data'
            'and xdata and ydata as the first arguments', DeprecationWarning)
        data = x
        x = kwargs.pop('xdata', 'x')
        y = kwargs.pop('ydata', 'y')

    plot_data = process_data_arguments(data=data,
                                       x=x,
                                       y=y,
                                       same_length=True,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if plot_data.distinct_datasets('x') == 1:
        default_legend_label = plot_data.get_keys('y')
    else:
        default_legend_label = plot_data.get_keys('x')
    if set_default_legend:
        plot_params.set_defaults(default_type='function', legend_label=default_legend_label)

    plot_params.set_defaults(default_type='function', name=default_legend_label)
    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    #Process the given color arguments
    plot_params.set_color_palette_by_num_plots()

    plot_kwargs = plot_params.plot_kwargs(plot_type='scatter')

    for indx, ((entry, source), plot_kw) in enumerate(zip(plot_data.items(), plot_kwargs)):

        res = p.scatter(x=entry.x, y=entry.y, source=source, **plot_kw, **kwargs)
        plot_params.add_tooltips(p, res, entry)

        if plot_params[('level', indx)] is not None:
            res.level = plot_params[('level', indx)]

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.set_legend(p)
    plot_params.save_plot(p, saveas)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_line(x,
               y=None,
               *,
               data=None,
               figure=None,
               xlabel='x',
               ylabel='y',
               title='',
               saveas='line',
               plot_points=False,
               area_curve=0,
               copy_data=False,
               set_default_legend=True,
               **kwargs):
    """
    Create an interactive multi-line plot with bokeh

    :param x: arraylike or key for data for the x-axis
    :param y: arraylike or key for data for the y-axis
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param xlabel: label for the x-axis
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure
    :param outfilename: filename of the output file
    :param plot_points: bool, if True also plot the points with a scatterplot on top
    :param copy_data: bool, if True the data argument will be copied
    :param set_default_legend: bool if True the data names are used to generate default legend labels

    Kwargs will be passed on to :py:class:`masci_tools.vis.bokeh_plotter.BokehPlotter`.
    If the arguments are not recognized they are passed on to the bokeh function `line`
    """
    from bokeh.models import ColumnDataSource

    if isinstance(x, (dict, pd.DataFrame, ColumnDataSource)) or x is None:
        warnings.warn(
            'Passing the source as first argument is deprecated. Please pass in source by the keyword data'
            'and xdata and ydata as the first arguments', DeprecationWarning)
        data = x
        x = kwargs.pop('xdata', 'x')
        y = kwargs.pop('ydata', 'y')

    plot_data = process_data_arguments(data=data,
                                       x=x,
                                       y=y,
                                       shift=area_curve,
                                       same_length=True,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if plot_data.distinct_datasets('x') == 1:
        default_legend_label = plot_data.get_keys('y')
    else:
        default_legend_label = plot_data.get_keys('x')
    if set_default_legend:
        plot_params.set_defaults(default_type='function', legend_label=default_legend_label)

    plot_params.set_defaults(default_type='function', name=default_legend_label)

    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    #Process the given color arguments
    plot_params.set_color_palette_by_num_plots()

    plot_kw_line = plot_params.plot_kwargs(plot_type='line')
    plot_kw_scatter = plot_params.plot_kwargs(plot_type='scatter')
    plot_kw_area = plot_params.plot_kwargs(plot_type='area')

    area_curve = kwargs.pop('area_curve', None)

    for indx, ((entry, source), kw_line, kw_scatter,
               kw_area) in enumerate(zip(plot_data.items(), plot_kw_line, plot_kw_scatter, plot_kw_area)):

        if plot_params[('area_plot', indx)]:
            if plot_params[('area_vertical', indx)]:
                p.harea(y=entry.y, x1=entry.x, x2=entry.shift, **kw_area, source=source)
            else:
                p.varea(x=entry.x, y1=entry.y, y2=entry.shift, **kw_area, source=source)

        res = p.line(x=entry.x, y=entry.y, source=source, **kw_line, **kwargs)
        plot_params.add_tooltips(p, res, entry)
        res2 = None
        if plot_points:
            res2 = p.scatter(x=entry.x, y=entry.y, source=source, **kw_scatter)

        if plot_params[('level', indx)] is not None:
            res.level = plot_params[('level', indx)]
            if res2 is not None:
                res2.level = plot_params[('level', indx)]

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.set_legend(p)
    plot_params.save_plot(p, saveas)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_dos(energy_grid,
              dos_data=None,
              *,
              data=None,
              energy_label='$$E-E_F [eV]$$',
              dos_label=r'DOS [1/eV]',
              title=r'Density of states',
              xyswitch=False,
              e_fermi=0,
              saveas='dos_plot',
              copy_data=False,
              **kwargs):
    """
    Create an interactive dos plot (non-spinpolarized) with bokeh
    Both horizontal or vertical orientation are possible

    :param energy_grid: arraylike or key data for the energy grid
    :param spin_up_data: arraylike or key data for the DOS
    :param data: source for the DOS data (optional) of the plot (pandas Dataframe for example)
    :param energy_label: label for the energy-axis
    :param dos_label: label for the dos-axis
    :param title: title of the figure
    :param xyswitch: bool if True, the energy will be plotted along the y-direction
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param outfilename: filename of the output file
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:func:`bokeh_line()`
    """
    from bokeh.models import ColumnDataSource

    if isinstance(energy_grid, (dict, pd.DataFrame, ColumnDataSource)) or energy_grid is None:
        warnings.warn(
            'Passing the dataframe as first argument is deprecated. Please pass in source by the keyword data'
            'and energy_grid and dos_data as the first arguments', DeprecationWarning)
        data = energy_grid
        energy_grid = kwargs.pop('energy', 'energy_grid')
        dos_data = kwargs.pop('ynames', None)

    if dos_data is None and data is not None:
        dos_data = set(data.keys()) - set([energy_grid] if isinstance(energy_grid, str) else energy_grid)
        dos_data = sorted(dos_data)

    plot_data = process_data_arguments(data=data,
                                       energy=energy_grid,
                                       dos=dos_data,
                                       same_length=True,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if 'limits' in kwargs:
        limits = kwargs.pop('limits')
        if 'x' not in limits and 'y' not in limits:
            if xyswitch:
                limits['x'], limits['y'] = limits.pop('dos', None), limits.pop('energy', None)
            else:
                limits['x'], limits['y'] = limits.pop('energy', None), limits.pop('dos', None)
        kwargs['limits'] = {k: v for k, v in limits.items() if v is not None}

    lines = {'horizontal': 0}
    lines['vertical'] = e_fermi

    if xyswitch:
        lines['vertical'], lines['horizontal'] = lines['horizontal'], lines['vertical']

    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             tooltips=[('Name', '$name'), ('Energy', '@{x}{{0.0[00]}}'),
                                       ('DOS value', '@$name{{0.00}}')],
                             figure_kwargs={
                                 'width': 1000,
                             })

    if xyswitch:
        x, y = plot_data.get_keys('dos'), plot_data.get_keys('energy')
        xlabel, ylabel = dos_label, energy_label
        plot_params.set_defaults(default_type='function', area_vertical=True)
    else:
        xlabel, ylabel = energy_label, dos_label
        x, y = plot_data.get_keys('energy'), plot_data.get_keys('dos')

    p = bokeh_line(x,
                   y,
                   data=plot_data.data,
                   xlabel=xlabel,
                   ylabel=ylabel,
                   title=title,
                   name=y,
                   saveas=saveas,
                   **kwargs)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_spinpol_dos(energy_grid,
                      spin_up_data=None,
                      spin_dn_data=None,
                      *,
                      data=None,
                      spin_dn_negative=True,
                      energy_label='$$E-E_F [eV]$$',
                      dos_label=r'DOS [1/eV]',
                      title=r'Density of states',
                      xyswitch=False,
                      e_fermi=0,
                      spin_arrows=True,
                      saveas='dos_plot',
                      copy_data=False,
                      **kwargs):
    """
    Create an interactive dos plot (spinpolarized) with bokeh
    Both horizontal or vertical orientation are possible

    :param energy_grid: arraylike or key data for the energy grid
    :param spin_up_data: arraylike or key data for the DOS spin-up
    :param spin_dn_data: arraylike or key data for the DOS spin-dn
    :param data: source for the DOS data (optional) of the plot (pandas Dataframe for example)
    :param spin_dn_negative: bool, if True (default), the spin down components are plotted downwards
    :param energy_label: label for the energy-axis
    :param dos_label: label for the dos-axis
    :param title: title of the figure
    :param xyswitch: bool if True, the energy will be plotted along the y-direction
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param spin_arrows: bool, if True (default) small arrows will be plotted on the left side of the plot indicating
                        the spin directions (if spin_dn_negative is True)
    :param outfilename: filename of the output file
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:func:`bokeh_line()`
    """
    from bokeh.models import NumeralTickFormatter, Arrow, NormalHead
    from bokeh.models import ColumnDataSource

    if isinstance(energy_grid, (dict, pd.DataFrame, ColumnDataSource)) or energy_grid is None:
        warnings.warn(
            'Passing the dataframe as first argument is deprecated. Please pass in source by the keyword data'
            'and energy_grid and dos_data as the first arguments', DeprecationWarning)
        data = energy_grid
        energy_grid = kwargs.pop('energy', 'energy_grid')
        spin_up_data = kwargs.pop('ynames', None)
        spin_up_data, spin_dn_data = spin_up_data[:len(spin_up_data) // 2], spin_up_data[len(spin_up_data) // 2:]

    if spin_up_data is None and data is not None:
        spin_up_data = {key for key in data.keys() if '_up' in key}
        spin_up_data = sorted(spin_up_data)
        spin_dn_data = {key for key in data.keys() if '_dn' in key}
        spin_dn_data = sorted(spin_dn_data)

    plot_data = process_data_arguments(data=data,
                                       energy=energy_grid,
                                       spin_up=spin_up_data,
                                       spin_dn=spin_dn_data,
                                       same_length=True,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if 'limits' in kwargs:
        limits = kwargs.pop('limits')
        if 'x' not in limits and 'y' not in limits:
            if xyswitch:
                limits['x'], limits['y'] = limits.pop('dos', None), limits.pop('energy', None)
            else:
                limits['x'], limits['y'] = limits.pop('energy', None), limits.pop('dos', None)
        kwargs['limits'] = {k: v for k, v in limits.items() if v is not None}

    lines = {'horizontal': 0}
    lines['vertical'] = e_fermi

    if spin_dn_negative:
        plot_data.apply('spin_dn', lambda x: -x)

    if xyswitch:
        lines['vertical'], lines['horizontal'] = lines['horizontal'], lines['vertical']

    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             tooltips=[('DOS Name', '$name'), ('Energy', '@{x}{{0.0[00]}}'),
                                       ('Value', '@$name{{(0,0.00)}}')],
                             figure_kwargs={'width': 1000})

    #Create the full data for the scatterplot
    energy_entries = plot_data.get_keys('energy') * 2
    dos_entries = plot_data.get_keys('spin_up') + plot_data.get_keys('spin_dn')
    sources = plot_data.data
    if isinstance(sources, list):
        sources = sources * 2

    if xyswitch:
        x, y = dos_entries, energy_entries
        xlabel, ylabel = dos_label, energy_label
        plot_params.set_defaults(default_type='function',
                                 area_vertical=True,
                                 x_axis_formatter=NumeralTickFormatter(format='(0,0)'))
    else:
        xlabel, ylabel = energy_label, dos_label
        x, y = energy_entries, dos_entries
        plot_params.set_defaults(default_type='function',
                                 area_vertical=True,
                                 y_axis_formatter=NumeralTickFormatter(format='(0,0)'))

    plot_params.set_parameters(color=kwargs.pop('color', None), color_palette=kwargs.pop('color_palette', None))
    plot_params.set_color_palette_by_num_plots()

    #Double the colors for spin up and down
    kwargs['color'] = list(plot_params['color']).copy()
    kwargs['color'].extend(kwargs['color'])

    if 'legend_label' not in kwargs:
        kwargs['legend_label'] = dos_entries
    else:
        if isinstance(kwargs['legend_label'], list):
            if len(kwargs['legend_label']) == len(plot_data):
                kwargs['legend_label'].extend(kwargs['legend_label'])

    if 'show' in kwargs:
        plot_params.set_parameters(show=kwargs.pop('show'))
    if 'save_plots' in kwargs:
        plot_params.set_parameters(save_plots=kwargs.pop('save_plots'))

    with NestedPlotParameters(plot_params):
        p = bokeh_line(x,
                       y,
                       xlabel=xlabel,
                       ylabel=ylabel,
                       title=title,
                       data=sources,
                       name=dos_entries,
                       show=False,
                       save_plots=False,
                       **kwargs)

    if spin_arrows and spin_dn_negative:

        #These are hardcoded because the parameters are not
        #reused anywhere (for now)
        x_pos = 50
        length = 70
        pad = 30
        height = p.plot_height - 100
        alpha = 0.5

        p.add_layout(
            Arrow(x_start=x_pos,
                  x_end=x_pos,
                  y_start=height - pad - length,
                  y_end=height - pad,
                  start_units='screen',
                  end_units='screen',
                  line_width=2,
                  line_alpha=alpha,
                  end=NormalHead(line_width=2, size=10, fill_alpha=alpha, line_alpha=alpha)))
        p.add_layout(
            Arrow(x_start=x_pos,
                  x_end=x_pos,
                  y_start=pad + length,
                  y_end=pad,
                  start_units='screen',
                  end_units='screen',
                  line_width=2,
                  line_alpha=alpha,
                  end=NormalHead(line_width=2, size=10, fill_alpha=alpha, line_alpha=alpha)))

    plot_params.save_plot(p, saveas)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_bands(kpath,
                bands=None,
                *,
                data=None,
                size_data=None,
                color_data=None,
                xlabel='',
                ylabel='$$E-E_F [eV]$$',
                title='',
                special_kpoints=None,
                markersize_min=3.0,
                markersize_scaling=10.0,
                saveas='bands_plot',
                scale_color=True,
                separate_bands=False,
                line_plot=False,
                band_index=None,
                copy_data=False,
                **kwargs):
    """
    Create an interactive bandstructure plot (non-spinpolarized) with bokeh
    Can make a simple plot or weight the size and color of the points against a given weight

    :param kpath: arraylike or key data for the kpoint data
    :param bands: arraylike or key data for the eigenvalues
    :param size_data: arraylike or key data the weights to emphasize (optional)
    :param color_data: str or arraylike, data for the color values with a colormap (optional)
    :param data: source for the bands data (optional) of the plot (pandas Dataframe for example)
    :param xlabel: label for the x-axis (default no label)
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param special_kpoints: list of tuples (str, float), place vertical lines at the given values
                            and mark them on the x-axis with the given label
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param markersize_min: minimum value used in scaling points for weight
    :param markersize_scaling: factor used in scaling points for weight
    :param outfilename: filename of the output file
    :param scale_color: bool, if True (default) the weight will be additionally shown via a colormapping
    :param line_plot: bool, if True the bandstructure will be plotted with lines
                      Here no weights are supported
    :param separate_bands: bool, if True the bandstructure will be separately plotted for each band
                           allows more specific parametrization
    :param band_index: data for which eigenvalue belongs to which band (needed for line_plot and separate_bands)
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:func:`bokeh_multi_scatter()` or :py:func:`bokeh_line()`
    """
    from bokeh.transform import linear_cmap
    from bokeh.models import ColumnDataSource

    if 'size_scaling' in kwargs:
        warnings.warn('size_scaling is deprecated. Use markersize_scaling instead', DeprecationWarning)
        markersize_scaling = kwargs.pop('size_scaling')

    if 'size_min' in kwargs:
        warnings.warn('size_min is deprecated. Use markersize_min instead', DeprecationWarning)
        markersize_min = kwargs.pop('size_min')

    if isinstance(kpath, (dict, pd.DataFrame, ColumnDataSource)) or kpath is None:
        warnings.warn(
            'Passing the dataframe as first argument is deprecated. Please pass in source by the keyword data'
            'and kpath and bands as the first arguments', DeprecationWarning)
        data = kpath
        kpath = kwargs.pop('k_label', 'kpath')
        bands = kwargs.pop('eigenvalues', 'eigenvalues_up')

    if 'weight' in kwargs:
        warnings.warn('The weight argument is deprecated. Use size_data and color_data instead', DeprecationWarning)
        size_data = kwargs.pop('weight')

    plot_data = process_data_arguments(single_plot=True,
                                       data=data,
                                       kpath=kpath,
                                       bands=bands,
                                       size=size_data,
                                       color=color_data,
                                       band_index=band_index,
                                       copy_data=copy_data,
                                       use_column_source=True)

    if line_plot and size_data is not None:
        raise ValueError('Bandstructure with lines and size scaling not supported')

    if line_plot and color_data is not None:
        raise ValueError('Bandstructure with lines and color mapping not supported')

    if line_plot or separate_bands:
        if band_index is None:
            raise ValueError('The data for band indices are needed for separate_bands and line_plot')
        plot_data.group_data('band_index')
        plot_data.sort_data('kpath')

    if scale_color and size_data is not None:
        if color_data is not None:
            raise ValueError('color_data should not be provided when scale_color is True')
        plot_data.copy_data('size', 'color', rename_original=True)

    if color_data is not None:
        kwargs['color'] = plot_data.get_keys('color')

    entries = plot_data.keys(first=True)
    if entries.size is not None:
        ylimits = (-15, 15)
        if 'limits' in kwargs:
            if 'y' in kwargs['limits']:
                ylimits = kwargs['limits']['y']

        data = plot_data.values(first=True)
        mask = np.logical_and(data.bands > ylimits[0], data.bands < ylimits[1])

        weight_max = plot_data.max('size', mask=mask)

        plot_params.set_defaults(default_type='function', marker_size=entries.size)
        if scale_color:
            plot_params.set_defaults(default_type='function',
                                     color=linear_cmap(entries.color, 'Blues256', weight_max, -0.05))

        transform = lambda size: markersize_min + markersize_scaling * size / weight_max
        plot_data.apply('size', transform)
    else:
        plot_params.set_defaults(default_type='function', color='black')

    if special_kpoints is None:
        special_kpoints = []

    xticks = []
    xticklabels = {}
    for label, pos in special_kpoints:
        if label in ('Gamma', 'g'):
            label = r'$$\Gamma$$'
        if pos.is_integer():
            xticklabels[int(pos)] = label
        xticklabels[pos] = label
        xticks.append(pos)

    lines = {'horizontal': 0}
    lines['vertical'] = xticks

    limits = {'y': (-15, 15)}
    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             x_ticks=xticks,
                             x_ticklabels_overwrite=xticklabels,
                             figure_kwargs={
                                 'width': 1280,
                                 'height': 720
                             },
                             x_range_padding=0.0,
                             y_range_padding=0.0,
                             legend_label='Eigenvalues',
                             limits=limits)

    if line_plot:
        return bokeh_line(plot_data.get_keys('kpath'),
                          plot_data.get_keys('bands'),
                          data=plot_data.data,
                          xlabel=xlabel,
                          ylabel=ylabel,
                          title=title,
                          set_default_legend=False,
                          saveas=saveas,
                          **kwargs)
    return bokeh_multi_scatter(plot_data.get_keys('kpath'),
                               plot_data.get_keys('bands'),
                               data=plot_data.data,
                               xlabel=xlabel,
                               ylabel=ylabel,
                               title=title,
                               set_default_legend=False,
                               saveas=saveas,
                               **kwargs)


@ensure_plotter_consistency(plot_params)
def bokeh_spinpol_bands(kpath,
                        bands_up=None,
                        bands_dn=None,
                        *,
                        size_data=None,
                        color_data=None,
                        data=None,
                        xlabel='',
                        ylabel='$$E-E_F [eV]$$',
                        title='',
                        special_kpoints=None,
                        markersize_min=3.0,
                        markersize_scaling=10.0,
                        saveas='bands_plot',
                        scale_color=True,
                        line_plot=False,
                        separate_bands=False,
                        band_index=None,
                        copy_data=False,
                        **kwargs):
    """
    Create an interactive bandstructure plot (spinpolarized) with bokeh
    Can make a simple plot or weight the size and color of the points against a given weight

    :param kpath: arraylike or key data for the kpoint data
    :param bands_up: arraylike or key data for the eigenvalues spin-up
    :param bands_dn: arraylike or key data for the eigenvalues spin-dn
    :param size_data: arraylike or key data the weights to emphasize (optional)
    :param color_data: str or arraylike, data for the color values with a colormap (optional)
    :param data: source for the bands data (optional) of the plot (pandas Dataframe for example)
    :param xlabel: label for the x-axis (default no label)
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param special_kpoints: list of tuples (str, float), place vertical lines at the given values
                            and mark them on the x-axis with the given label
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param markersize_min: minimum value used in scaling points for weight
    :param markersize_scaling: factor used in scaling points for weight
    :param outfilename: filename of the output file
    :param scale_color: bool, if True (default) the weight will be additionally shown via a colormapping
    :param line_plot: bool, if True the bandstructure will be plotted with lines
                      Here no weights are supported
    :param separate_bands: bool, if True the bandstructure will be separately plotted for each band
                           allows more specific parametrization
    :param band_index: data for which eigenvalue belongs to which band (needed for line_plot and separate_bands)
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:func:`bokeh_multi_scatter()` or :py:func:`bokeh_line()`
    """
    from bokeh.transform import linear_cmap
    from bokeh.models import ColumnDataSource

    if 'size_scaling' in kwargs:
        warnings.warn('size_scaling is deprecated. Use markersize_scaling instead', DeprecationWarning)
        markersize_scaling = kwargs.pop('size_scaling')

    if 'size_min' in kwargs:
        warnings.warn('size_min is deprecated. Use markersize_min instead', DeprecationWarning)
        markersize_min = kwargs.pop('size_min')

    if isinstance(kpath, (dict, pd.DataFrame, ColumnDataSource)) or kpath is None:
        warnings.warn(
            'Passing the dataframe as first argument is deprecated. Please pass in source by the keyword data'
            'and kpath and bands_up and bands_dn as the first arguments', DeprecationWarning)
        data = kpath
        kpath = kwargs.pop('k_label', 'kpath')
        bands_up = kwargs.pop('eigenvalues', ['eigenvalues_up', 'eigenvalues_down'])
        bands_up, bands_dn = bands_up[0], bands_up[1]

    if 'weight' in kwargs:
        warnings.warn('The weight argument is deprecated. Use size_data and color_data instead', DeprecationWarning)
        size_data = kwargs.pop('weight')

    plot_data = process_data_arguments(data=data,
                                       kpath=kpath,
                                       bands=[bands_up, bands_dn],
                                       size=size_data,
                                       color=color_data,
                                       band_index=band_index,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if len(plot_data) != 2:
        raise ValueError('Wrong number of plots specified (Only 2 permitted)')

    if line_plot and size_data is not None:
        raise ValueError('Bandstructure with lines and size scaling not supported')

    if line_plot and color_data is not None:
        raise ValueError('Bandstructure with lines and color mapping not supported')

    if line_plot or separate_bands:
        if band_index is None:
            raise ValueError('The data for band indices are needed for separate_bands and line_plot')

        plot_data.group_data('band_index')
        plot_data.sort_data('kpath')

    if scale_color and size_data is not None:
        if color_data is not None:
            raise ValueError('color_data should not be provided when scale_color is True')
        plot_data.copy_data('size', 'color', rename_original=True)

    if color_data is not None:
        kwargs['color'] = plot_data.get_keys('color')

    if any(entry.size is not None for entry in plot_data.keys()):

        ylimits = (-15, 15)
        if 'limits' in kwargs:
            if 'y' in kwargs['limits']:
                ylimits = kwargs['limits']['y']

        data = plot_data.values()
        mask = [np.logical_and(col.bands > ylimits[0], col.bands < ylimits[1]) for col in data]
        weight_max = plot_data.max('size', mask=mask)

        transform = lambda size: markersize_min + markersize_scaling * size / weight_max
        plot_data.apply('size', transform)

        plot_params.set_defaults(default_type='function', marker_size=plot_data.get_keys('size'))
        if scale_color:
            plot_params.set_defaults(default_type='function',
                                     color=[
                                         linear_cmap(name, palette, weight_max, -0.05)
                                         for name, palette in zip(plot_data.get_keys('color'), ['Blues256', 'Reds256'])
                                     ])
    else:
        color = ['blue', 'red']
        plot_params.set_defaults(default_type='function', color=color)

    if special_kpoints is None:
        special_kpoints = []

    xticks = []
    xticklabels = {}
    for label, pos in special_kpoints:
        if label in ('Gamma', 'g'):
            label = r'$$\Gamma$$'
        if pos.is_integer():
            xticklabels[int(pos)] = label
        xticklabels[pos] = label
        xticks.append(pos)

    lines = {'horizontal': 0}
    lines['vertical'] = xticks

    limits = {'y': (-15, 15)}
    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             x_ticks=xticks,
                             x_ticklabels_overwrite=xticklabels,
                             figure_kwargs={
                                 'width': 1280,
                                 'height': 720
                             },
                             x_range_padding=0.0,
                             y_range_padding=0.0,
                             limits=limits,
                             legend_label=['Spin Up', 'Spin Down'],
                             level=[None, 'underlay'])

    if line_plot or separate_bands:
        plot_params.num_plots = len(plot_data)
        kwargs = plot_params.expand_parameters(original_length=2, **kwargs)

    if line_plot:
        return bokeh_line(plot_data.get_keys('kpath'),
                          plot_data.get_keys('bands'),
                          data=plot_data.data,
                          xlabel=xlabel,
                          ylabel=ylabel,
                          title=title,
                          set_default_legend=False,
                          saveas=saveas,
                          **kwargs)
    return bokeh_multi_scatter(plot_data.get_keys('kpath'),
                               plot_data.get_keys('bands'),
                               data=plot_data.data,
                               xlabel=xlabel,
                               ylabel=ylabel,
                               title=title,
                               set_default_legend=False,
                               saveas=saveas,
                               **kwargs)


@ensure_plotter_consistency(plot_params)
def bokeh_spectral_function(kpath,
                            energy_grid,
                            spectral_function,
                            *,
                            data=None,
                            special_kpoints=None,
                            e_fermi=0,
                            xlabel='',
                            ylabel='$$E-E_F [eV]$$',
                            title='',
                            saveas='spectral_function',
                            copy_data=False,
                            figure=None,
                            **kwargs):
    """
    Create a colormesh plot of a spectral function

    :param kpath: data for the kpoint coordinates
    :param energy_grid: data for the energy grid
    :param spectral_function: 2D data for the spectral function
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param title: str, Title of the plot
    :param xlabel: str, label for the x-axis
    :param ylabel: str, label for the y-axis
    :param saveas: str, filename for the saved plot
    :param e_fermi: float (default 0), place the line for the fermi energy at this value
    :param special_kpoints: list of tuples (str, float), place vertical lines at the given values
                            and mark them on the x-axis with the given label
    :param copy_data: bool, if True the data argument will be copied

    All other Kwargs are passed on to the image call of bokeh
    """

    plot_data = process_data_arguments(single_plot=True,
                                       data=data,
                                       kpath=kpath,
                                       energy=energy_grid,
                                       spectral_function=spectral_function,
                                       forbid_split_up={
                                           'spectral_function',
                                       },
                                       copy_data=copy_data)

    if special_kpoints is None:
        special_kpoints = []

    xticks = []
    xticklabels = {}
    for label, pos in special_kpoints:
        if label in ('Gamma', 'g'):
            label = r'$$\Gamma$$'
        if pos.is_integer():
            xticklabels[int(pos)] = label
        xticklabels[pos] = label
        xticks.append(pos)

    lines = {'horizontal': e_fermi}
    lines['vertical'] = xticks

    limits = {'y': (plot_data.min('energy'), plot_data.max('energy'))}
    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             x_ticks=xticks,
                             x_ticklabels_overwrite=xticklabels,
                             figure_kwargs={
                                 'width': 1280,
                                 'height': 720
                             },
                             x_range_padding=0.0,
                             y_range_padding=0.0,
                             limits=limits,
                             color_palette='Plasma256',
                             legend_label='Spectral function',
                             straight_line_options={'line_color': 'white'})

    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    entry = plot_data.values(first=True)

    plot_kw = plot_params.plot_kwargs(plot_type='image')

    min_energy = plot_data.min('energy')
    dh = plot_data.max('energy') - min_energy
    p.image([entry.spectral_function],
            x=0,
            y=plot_data.min('energy'),
            dh=dh,
            dw=plot_data.max('kpath'),
            **plot_kw,
            **kwargs)

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.set_legend(p)
    plot_params.save_plot(p, saveas)

    return p


####################################################################################################
##################################### special plots ################################################
####################################################################################################


@ensure_plotter_consistency(plot_params)
def periodic_table_plot(
        values,
        positions=None,
        *,
        color_data=None,
        log_scale=False,
        color_map=None,
        data=None,
        copy_data=False,
        title='',
        saveas='periodictable.html',
        blank_outsiders='both',  #min, max or both, None
        blank_color='#c4c4c4',
        include_legend=True,
        figure=None,
        **kwargs):
    """
    Plot function for an interactive periodic table plot. Heat map and hover tool.
    source must be a pandas dataframe containing, atom period and group, atomic number and symbol

    :param values: data for the text inside each elements box
    :param positions: y positions relative to the middle of the box for each value
    :param color_data: data to display as a heatmap
    :param color_map: color palette to use for the heatmap (default matplotlib plasma)
    :param log_scale: bool, if True the heatmap is done logarithmically
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param title: str, Title of the plot
    :param saveas: str, filename for the saved plot
    :param blank_outsiders: either 'both', 'min', 'max' or None, determines, which points outside the color
                            range to color with a default blank color
    :param blank_color: color to replace values outside the color range by
    :param include_legend: if True an additional entry with labels explaing each value entry is added
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure

    Additional kwargs are passed on to the label creation for the element box
    The kwargs `legend_options` and `colorbar_options` can be used to overwrite default
    values for these regions of the plot
    """
    from matplotlib.colors import Normalize, LogNorm
    from matplotlib.cm import ScalarMappable
    from matplotlib.cm import plasma  #pylint: disable=no-name-in-module

    from bokeh.transform import dodge, linear_cmap, log_cmap
    from bokeh.sampledata.periodic_table import elements
    from bokeh.models import Label, ColorBar, OpenHead, Arrow, BasicTicker

    from bokeh.models import ColumnDataSource

    if isinstance(values, (dict, pd.DataFrame, ColumnDataSource)) or values is None:
        warnings.warn(
            'Passing the dataframe as first argument is deprecated. Please pass in source by the keyword data'
            'and values and positions as the first arguments', DeprecationWarning)
        data = values
        values = kwargs.pop('display_values', [])
        positions = kwargs.pop('display_positions', [])

    if 'color_value' in kwargs:
        warnings.warn('color_value is deprecated. Use color_data instead', DeprecationWarning)
        color_data = kwargs.pop('color_value')

    if 'outfilename' in kwargs:
        warnings.warn('outfilename is deprecated. Use saveas instead', DeprecationWarning)
        saveas = kwargs.pop('outfilename')

    if 'bokeh_palette' in kwargs:
        warnings.warn('bokeh_palette is deprecated. Use color_palette instead', DeprecationWarning)
        kwargs['color_palette'] = kwargs.pop('bokeh_palette')

    if 'copy_source' in kwargs:
        warnings.warn('copy_source is deprecated. Use copy_data instead', DeprecationWarning)
        copy_data = kwargs.pop('copy_source')

    if 'legend_labels' in kwargs:
        warnings.warn('legend_labels is deprecated. Use legend_label instead', DeprecationWarning)
        kwargs['legend_label'] = kwargs.pop('legend_labels')

    if 'color_bar_title' in kwargs:
        warnings.warn('color_bar_title is deprecated. Use title entry in the colorbar_options argument instead',
                      DeprecationWarning)
        kwargs.setdefault('colorbar_options', {})['title'] = kwargs.pop('color_bar_title')

    if 'value_color_range' in kwargs:
        warnings.warn('The value_color_range argument is deprecated. Use the color key in the limits argument instead',
                      DeprecationWarning)
        kwargs.setdefault('limits', {})['color'] = kwargs.pop('value_color_range')

    if not isinstance(blank_outsiders, str):
        warnings.warn(
            'The blank_outsiders argument as a list of bools is deprecated. Use min, max or both or None instead',
            DeprecationWarning)
        if all(blank_outsiders):
            blank_outsiders = 'both'
        elif blank_outsiders[0]:
            blank_outsiders = 'min'
        elif blank_outsiders[1]:
            blank_outsiders = 'max'
        else:
            blank_outsiders = None

    if color_map is None:
        color_map = plasma

    #For this plot we use the sample data from bokeh to fill in values
    if data is None:
        data = elements

    if positions is None:
        raise NotImplementedError('Not providing positions is not yet implemented')

    if isinstance(color_data, list):
        raise ValueError('Only one color data entry allowed')

    plot_data = process_data_arguments(data=data,
                                       copy_data=copy_data,
                                       color=color_data,
                                       values=values,
                                       forbid_split_up={'color'})

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    #Create two dictionary parameters for customizing the legend and colorbar
    plot_params.add_parameter(
        'legend_options',
        default_val={
            'text_font_size': '13px',  #Please only provide it in pixels
            'arrow_line_width': 2,
            'arrow_size': 4,
            'arrow_length': 0.3,
            'label_standoff': 0.0,
        })

    plot_params.add_parameter('colorbar_options',
                              default_val={
                                  'height': 40,
                                  'width': 500,
                                  'fontsize': 12,
                                  'label_standoff': 8,
                                  'title': plot_data.keys(first=True).color,
                                  'scale_alpha': 1.0
                              })

    groups = [str(x) for x in range(1, 19)]
    periods = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII']
    data['period'] = [periods[x - 1] for x in data.period]
    plot_params.set_defaults(default_type='function',
                             figure_kwargs={
                                 'width': 1475,
                                 'height': 675,
                                 'x_range': groups,
                                 'y_range': list(reversed(periods)),
                             },
                             color_palette='Plasma256',
                             format_tooltips=False,
                             tooltips=[('Name', '@name'), ('Atomic number', '@{atomic number}'),
                                       ('Atomic mass', '@{atomic mass}'), ('CPK color', '$color[hex, swatch]:CPK'),
                                       ('Electronic configuration', '@{electronic configuration}')])

    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title, '', '', figure=figure)

    color_scale = None
    if any(entry.color is not None for entry in plot_data.keys()):
        if plot_params['limits'] is not None and plot_params['limits'].get('color') is not None:
            min_color, max_color = plot_params['limits']['color']
        else:
            min_color, max_color = plot_data.min('color'), plot_data.max('color')
        color_values = plot_data.values(first=True).color
        color_name = plot_data.keys(first=True).color

        if not log_scale:
            color_mapper = linear_cmap(color_name, palette=plot_params['color_palette'], low=min_color, high=max_color)
            norm = Normalize(vmin=min_color, vmax=max_color)
        else:
            if min_color < 0:
                raise ValueError(f"Entry for 'color' element '{color_data}' is negative but log-scale is selected")
            color_mapper = log_cmap(color_name, palette=plot_params['color_palette'], low=min_color, high=max_color)
            norm = LogNorm(vmin=min_color, vmax=max_color)

        color_scale = ScalarMappable(norm=norm, cmap=color_map).to_rgba(color_values, alpha=None)
        plot_params.set_defaults(default_type='function', color=color_mapper)

        if blank_outsiders is not None:
            if blank_outsiders == 'both':
                outsiders = np.logical_or(color_values < min_color, color_values > max_color)
            elif blank_outsiders == 'min':
                outsiders = color_values < min_color
            elif blank_outsiders == 'max':
                outsiders = color_values > max_color
            plot_data.mask_data(outsiders, data_key='color', replace_value=blank_color)

    if include_legend:
        # we copy the Be entry and display it with some text again at another spot
        be = data[3:4].copy()
        be['group'] = '7'
        data.loc[-1] = be.values[0]
        data.index = data.index + 1
        data = data.sort_index()

    r = p.rect(
        'group',
        'period',
        0.95,
        0.95,
        source=data,
        fill_alpha=0.6,  # legend="metal",
        color=plot_params['color'])
    plot_params.add_tooltips(p, r)

    text_props = {'source': data, 'text_align': 'left', 'text_baseline': 'middle'}
    x = dodge('group', -0.4, range=p.x_range)
    # The element names
    p.text(x=x,
           y=dodge('period', -0.35, range=p.y_range),
           text='symbol',
           text_font_style='bold',
           text_font_size='14pt',
           **text_props)
    p.text(x=dodge('group', 0.18, range=p.x_range),
           y=dodge('period', 0.3, range=p.y_range),
           text='atomic number',
           text_font_size='12pt',
           **text_props)

    plot_kw = plot_params.plot_kwargs()

    # The values displayed on the element boxes
    for entry, kw, position in zip(plot_data.keys(), plot_kw, positions):
        p.text(x=x,
               y=dodge('period', position, range=p.y_range),
               text=entry.values,
               text_font_size='10pt',
               **text_props)

        label = kw.pop('legend_label', entry.values)

        if include_legend:
            options = plot_params['legend_options']
            arrow_length = options['arrow_length']
            y_pos = 5.5 + position
            x_pos = 7 + options['label_standoff'] + arrow_length
            legend_fontsize = options['text_font_size']
            fontsize_in_data = int(legend_fontsize.rstrip('px')) / plot_params['figure_kwargs']['height'] * 7
            y_pos_label = y_pos - fontsize_in_data / 2

            # legend
            # not a real legend, but selfmade text
            # I do not like the hardcoded positions of the legend
            legendlabel = Label(x=x_pos,
                                y=y_pos_label,
                                text=label,
                                render_mode='canvas',
                                border_line_color='black',
                                border_line_alpha=0.0,
                                background_fill_color=None,
                                background_fill_alpha=1.0,
                                text_font_size=legend_fontsize)
            p.add_layout(legendlabel)

            legendlabelarrow = Arrow(x_start=x_pos,
                                     x_end=x_pos - arrow_length,
                                     y_start=y_pos,
                                     y_end=y_pos,
                                     line_width=options['arrow_line_width'],
                                     end=OpenHead(line_width=options['arrow_line_width'], size=options['arrow_size']))
            p.add_layout(legendlabelarrow)

    p.yaxis.major_label_text_font_size = '22pt'
    p.xaxis.visible = False
    p.yaxis.visible = False
    p.outline_line_color = None
    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_standoff = 0
    # p.legend.orientation = "horizontal"

    # add color bar

    if any(entry.color is not None for entry in plot_data.keys()):
        colorbar_options = plot_params['colorbar_options'].copy()
        cbar_fontsize = f"{colorbar_options.pop('fontsize')}pt"
        cbar_location = (plot_params['figure_kwargs']['width'] * 0.2, plot_params['figure_kwargs']['height'] * 0.55)

        color_bar = ColorBar(color_mapper=color_mapper['transform'],
                             title_text_font_size='12pt',
                             ticker=BasicTicker(desired_num_ticks=10),
                             border_line_color=None,
                             background_fill_color=None,
                             location=cbar_location,
                             orientation='horizontal',
                             major_label_text_font_size=cbar_fontsize,
                             **colorbar_options)

        p.add_layout(color_bar, 'center')

    # deactivate grid
    p.grid.grid_line_color = None
    plot_params.set_limits(p)
    plot_params.save_plot(p, saveas)

    return p


@ensure_plotter_consistency(plot_params)
def plot_lattice_constant(scaling,
                          total_energy,
                          *,
                          fit_data=None,
                          data=None,
                          figure=None,
                          relative=True,
                          ref_const=None,
                          title='Equation of states',
                          saveas='lattice_constant',
                          copy_data=False,
                          **kwargs):
    """
    Plot a lattice constant versus Total energy
    Plot also the fit.
    On the x axis is the scaling, it

    :param scaling: arraylike, data for the scaling factor
    :param total_energy: arraylike, data for the total energy
    :param fit_data: arraylike, optional data of fitted data
    :param relative: bool, scaling factor given (True), or lattice constants given?
    :param ref_const: float (optional), or list of floats, lattice constant for scaling 1.0
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param copy_data: bool if True the data argument will be copied
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure

    Function specific parameters:
        :param marker_fit: defaults to `marker`, marker type for the fit data
        :param marker_size_fit: defaults to `marker_size`, markersize for the fit data
        :param line_width_fit: defaults to `line_width`, linewidth for the fit data
        :param legend_label_fit: str label for the fit data

    Other Kwargs will be passed on to :py:func:`bokeh_line()`
    """
    # TODO: make box which shows fit results. (fit resuls have to be past)

    plot_data = process_data_arguments(data=data,
                                       scaling=scaling,
                                       energy=total_energy,
                                       fit=fit_data,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if relative:
        if ref_const:
            xlabel = rf'Relative Volume [a/{ref_const}$\AA$]'
        else:
            xlabel = r'Relative Volume'
    else:
        xlabel = r'Volume [$\AA$]'

    if len(plot_data) > 1:
        ylabel = r'Total energy norm[0] [eV]'
    else:
        ylabel = r'Total energy [eV]'

    #Add custom parameters for fit
    plot_params.add_parameter('marker_fit', default_from='marker')
    plot_params.add_parameter('marker_size_fit', default_from='marker_size')
    plot_params.add_parameter('line_width_fit', default_from='line_width')
    plot_params.add_parameter('legend_label_fit')

    plot_params.set_defaults(default_type='function',
                             marker_fit='square',
                             legend_label='simulation data',
                             legend_label_fit='fit results',
                             color='black' if len(plot_data) == 1 else None)

    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title=title, xlabel=xlabel, ylabel=ylabel, figure=figure)

    plot_kw = plot_params.plot_kwargs(post_process=False)
    plot_fit_kw_line = plot_params.plot_kwargs(post_process=False,
                                               plot_type='line',
                                               line_width='line_width_fit',
                                               legend_label='legend_label_fit')
    plot_fit_kw_scatter = plot_params.plot_kwargs(post_process=False,
                                                  plot_type='scatter',
                                                  marker='marker_fit',
                                                  marker_size='marker_size_fit',
                                                  legend_label='legend_label_fit')
    plot_fit_kw = {**plot_fit_kw_line, **plot_fit_kw_scatter}

    with NestedPlotParameters(plot_params):
        p = bokeh_line(plot_data.get_keys('scaling'),
                       plot_data.get_keys('energy'),
                       data=plot_data.data,
                       xlabel=xlabel,
                       ylabel=ylabel,
                       title=title,
                       figure=p,
                       show=False,
                       save_plots=False,
                       plot_points=True,
                       **plot_kw,
                       **kwargs)
    if any(entry.fit is not None for entry in plot_data.keys()):
        with NestedPlotParameters(plot_params):
            p = bokeh_line(plot_data.get_keys('scaling'),
                           plot_data.get_keys('fit'),
                           data=plot_data.data,
                           xlabel=xlabel,
                           ylabel=ylabel,
                           title=title,
                           figure=p,
                           show=False,
                           save_plots=False,
                           plot_points=True,
                           **plot_fit_kw,
                           **kwargs)

    plot_params.draw_straight_lines(p)
    plot_params.save_plot(p, saveas)

    return p


######## a 2d matrix plot ##########

######### plot convergence results plot ########


@ensure_plotter_consistency(plot_params)
def plot_convergence(iteration,
                     distance,
                     total_energy,
                     *,
                     data=None,
                     saveas_energy='energy_convergence',
                     saveas_distance='distance_convergence',
                     figure_energy=None,
                     figure_distance=None,
                     xlabel='Iteration',
                     ylabel_energy='Total energy difference [Htr]',
                     ylabel_distance='Distance [me/bohr^3]',
                     title_energy='Total energy difference over scf-Iterations',
                     title_distance='Convergence (log)',
                     copy_data=False,
                     drop_last_iteration=False,
                     **kwargs):
    """
    Plot the total energy differences versus the scf iteration
    and plot the distance of the density versus iterations.

    :param iteration: data for the number of iterations
    :param distance: data of distances
    :param total_energy: data of total energies
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param xlabel: str, label for the x-axis of both plots
    :param saveas_energy: str, filename for the energy convergence plot
    :param figure_energy: Axes object for the energy convergence plot
    :param title_energy: str, title for the energy convergence plot
    :param ylabel_energy: str, label for the y-axis for the energy convergence plot
    :param saveas_distance: str, filename for the distance plot
    :param figure_distance: Axes object for the distance plot
    :param title_distance: str, title for the distance plot
    :param ylabel_distance: str, label for the y-axis for the distance plot
    :param copy_data: bool if  True the data argument is copied
    :param drop_last_iteration: bool if True the last iteration is dropped for the distance plot

    Other Kwargs will be passed on to all :py:func:`bokeh_line()` calls
    """

    plot_data = process_data_arguments(data=data,
                                       iteration=iteration,
                                       distance=distance,
                                       energy=total_energy,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    #Calculate energy differences and corresponding
    plot_data.copy_data('energy', 'energy_diff')
    plot_data.copy_data('iteration', 'iteration_energy')
    plot_data.apply('energy_diff', np.diff)
    plot_data.apply('energy_diff', np.abs)
    plot_data.apply('iteration_energy', np.delete, obj=0)
    plot_data.apply('iteration_energy', np.append, values=np.nan)
    plot_data.apply('energy_diff', np.append, values=np.nan)

    if drop_last_iteration:
        plot_data.apply('iteration', np.delete, obj=-1)

    if len(plot_data) == 1:
        default_energy_label = 'delta total energy'
        default_distance_label = 'distance'
    else:
        default_energy_label = [f'delta total energy {i}' for i in range(len(plot_data))]
        default_distance_label = [f'distance {i}' for i in range(len(plot_data))]

    plot_params.set_defaults(default_type='function',
                             legend_label=default_energy_label,
                             color='black' if len(plot_data) == 1 else None,
                             tooltips=[('Calculation id', '$name'), ('Iteration', '@{x}'),
                                       ('Total energy difference', '@{y}')],
                             figure_kwargs={
                                 'plot_width': 800,
                                 'plot_height': 450,
                                 'y_axis_type': 'log',
                                 'x_axis_type': 'linear',
                             },
                             legend_outside_plot_area=True)

    with NestedPlotParameters(plot_params):
        p1 = bokeh_line(plot_data.get_keys('iteration_energy'),
                        plot_data.get_keys('energy_diff'),
                        data=plot_data.data,
                        xlabel=xlabel,
                        ylabel=ylabel_energy,
                        title=title_energy,
                        saveas=saveas_energy,
                        figure=figure_energy,
                        plot_points=True,
                        set_default_legend=False,
                        **kwargs)

    plot_params.set_defaults(default_type='function',
                             legend_label=default_distance_label,
                             tooltips=[('Calculation id', '$name'), ('Iteration', '@{x}'), ('Charge distance', '@{y}')])

    with NestedPlotParameters(plot_params):
        p2 = bokeh_line(plot_data.get_keys('iteration'),
                        plot_data.get_keys('distance'),
                        data=plot_data.data,
                        xlabel=xlabel,
                        ylabel=ylabel_distance,
                        title=title_distance,
                        saveas=saveas_distance,
                        figure=figure_distance,
                        plot_points=True,
                        set_default_legend=False,
                        **kwargs)

    return p1, p2


@ensure_plotter_consistency(plot_params)
def plot_convergence_results(iteration, distance, total_energy, *, saveas='convergence', **kwargs):
    """
    Plot the total energy versus the scf iteration
    and plot the distance of the density versus iterations. Uses bokeh_line and bokeh_scatter

    :param iteration: list of Int
    :param distance: list of floats
    :total_energy: list of floats
    :param show: bool, if True call show

    Kwargs will be passed on to :py:func:`bokeh_line()`

    :returns grid: bokeh grid with figures
    """
    from bokeh.layouts import gridplot

    warnings.warn(
        'plot_convergence_results is deprecated. Use the more general plot_convergence instead.'
        'It can do both single and multiple calculations natively', DeprecationWarning)

    if 'show' in kwargs:
        plot_params.set_parameters(show=kwargs.pop('show'))
    if 'save_plots' in kwargs:
        plot_params.set_parameters(save_plots=kwargs.pop('save_plots'))

    with NestedPlotParameters(plot_params):
        p1, p2 = plot_convergence(iteration, distance, total_energy, save_plots=False, show=False, **kwargs)

    grid = gridplot([p1, p2], ncols=1)

    plot_params.save_plot(grid, saveas)

    return grid


@ensure_plotter_consistency(plot_params)
def plot_convergence_results_m(iterations,
                               distances,
                               total_energies,
                               *,
                               link=False,
                               nodes=None,
                               modes=None,
                               plot_label=None,
                               saveas='convergence',
                               **kwargs):
    """
    Plot the total energy versus the scf iteration
    and plot the distance of the density versus iterations in a bokeh grid for several SCF results.

    :param distances: list of lists of floats
    :total_energies: list of lists of floats
    :param iterations: list of lists of Int
    :param link: bool, optional default=False:
    :param nodes: list of node uuids or pks important for links
    :param saveas1: str, optional default='t_energy_convergence', save first figure as
    :param saveas2: str, optional default='distance_convergence', save second figure as
    :param figure_kwargs: dict, optional default={'plot_width': 600, 'plot_height': 450}, gets parsed
                          to bokeh_line
    :param kwargs: further key-word arguments for bokeh_line

    :returns grid: bokeh grid with figures
    """
    from bokeh.layouts import gridplot

    warnings.warn(
        'plot_convergence_results_m is deprecated. Use the more general plot_convergence instead.'
        'It can do both single and multiple calculations natively', DeprecationWarning)

    if 'show' in kwargs:
        plot_params.set_parameters(show=kwargs.pop('show'))
    if 'save_plots' in kwargs:
        plot_params.set_parameters(save_plots=kwargs.pop('save_plots'))
    if plot_label is not None:
        kwargs['legend_label'] = plot_label

    if modes is None:
        modes = []

    with NestedPlotParameters(plot_params):
        p1, p2 = plot_convergence(iterations,
                                  distances,
                                  total_energies,
                                  save_plots=False,
                                  show=False,
                                  drop_last_iteration=any(mode == 'force' for mode in modes),
                                  **kwargs)

    grid = gridplot([p1, p2], ncols=1)

    plot_params.save_plot(grid, saveas)

    return grid


@ensure_plotter_consistency(plot_params)
def matrix_plot(
        text_values,
        x_axis_data,
        y_axis_data,
        positions=None,
        *,
        color_data=None,
        secondary_color_data=None,
        x_offset=-0.47,
        log_scale=False,
        color_map=None,
        data=None,
        copy_data=False,
        title='',
        xlabel='x',
        ylabel='y',
        saveas='matrix_plot.html',
        blank_outsiders='both',  #min, max or both, None
        blank_color='#c4c4c4',
        figure=None,
        categorical_axis=False,
        categorical_sort_key=None,
        block_size=0.95,
        block_size_pixel=100,
        **kwargs):
    """
    Plot function for an interactive periodic table plot. Heat map and hover tool.
    source must be a pandas dataframe containing, atom period and group, atomic number and symbol

    :param values: data for the text inside each elements box
    :param positions: y positions relative to the middle of the box for each value
    :param color_data: data to display as a heatmap
    :param color_map: color palette to use for the heatmap (default matplotlib plasma)
    :param log_scale: bool, if True the heatmap is done logarithmically
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param title: str, Title of the plot
    :param saveas: str, filename for the saved plot
    :param blank_outsiders: either 'both', 'min', 'max' or None, determines, which points outside the color
                            range to color with a default blank color
    :param blank_color: color to replace values outside the color range by
    :param include_legend: if True an additional entry with labels explaing each value entry is added
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure

    Additional kwargs are passed on to the label creation for the element box
    The kwargs `legend_options` and `colorbar_options` can be used to overwrite default
    values for these regions of the plot
    """
    from matplotlib.cm import plasma  #pylint: disable=no-name-in-module

    from bokeh.transform import dodge, linear_cmap, log_cmap
    from bokeh.models import FactorRange, ColorBar, BasicTicker

    if color_map is None:
        color_map = plasma

    if positions is None:
        raise NotImplementedError('Not providing positions is not yet implemented')

    plot_data = process_data_arguments(data=data,
                                       copy_data=copy_data,
                                       color=color_data,
                                       secondary_color=secondary_color_data,
                                       text=text_values,
                                       x_axis=x_axis_data,
                                       y_axis=y_axis_data,
                                       forbid_split_up={'color', 'secondary_color', 'x_axis', 'y_axis'})

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    plot_params.add_parameter('colorbar_options',
                              default_val={
                                  'fontsize': 12,
                                  'label_standoff': 8,
                                  'title': plot_data.keys(first=True).color,
                                  'scale_alpha': 1.0
                              })

    plot_params.set_defaults(default_type='function',
                             figure_kwargs={
                                 'x_axis_type': 'auto',
                                 'y_axis_type': 'auto',
                             },
                             color_palette='Plasma256',
                             format_tooltips=False,
                             tooltips=[])

    if categorical_axis:
        x_values = sorted(set(plot_data.values(first=True).x_axis), key=categorical_sort_key)
        y_values = sorted(set(plot_data.values(first=True).y_axis), key=categorical_sort_key)
        plot_params.set_defaults(default_type='function',
                                 figure_kwargs={
                                     'height': block_size_pixel * len(y_values),
                                     'width': block_size_pixel * len(x_values),
                                     'x_range': FactorRange(factors=x_values),
                                     'y_range': FactorRange(factors=y_values),
                                 })

    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    if any(entry.color is not None for entry in plot_data.keys()):
        if plot_params['limits'] is not None and plot_params['limits'].get('color') is not None:
            min_color, max_color = plot_params['limits']['color']
        else:
            min_color, max_color = plot_data.min('color'), plot_data.max('color')
            if any(entry.secondary_color is not None for entry in plot_data.keys()):
                min_color = min(min_color, plot_data.min('secondary_color'))
                max_color = max(max_color, plot_data.max('secondary_color'))

        color_values = [plot_data.values(first=True).color]
        color_name = [plot_data.keys(first=True).color]
        if any(entry.secondary_color is not None for entry in plot_data.keys()):
            color_values += [plot_data.values(first=True).secondary_color]
            color_name += [plot_data.keys(first=True).secondary_color]

        color_mappers = []
        for name, value in zip(color_name, color_values):
            if not log_scale:
                color_mappers.append(
                    linear_cmap(name, palette=plot_params['color_palette'], low=min_color, high=max_color))
            else:
                if min_color < 0:
                    raise ValueError(f"Entry for 'color' element '{color_data}' is negative but log-scale is selected")
                color_mappers.append(log_cmap(name, palette=plot_params['color_palette'], low=min_color,
                                              high=max_color))

            if blank_outsiders is not None:
                if blank_outsiders == 'both':
                    outsiders = np.logical_or(value < min_color, value > max_color)
                elif blank_outsiders == 'min':
                    outsiders = value < min_color
                elif blank_outsiders == 'max':
                    outsiders = value > max_color
                plot_data.mask_data(outsiders, data_key='color', replace_value=blank_color)
        plot_params.set_defaults(default_type='function', color=color_mappers)

    entry, source = plot_data.items(first=True)
    if any(entry.secondary_color is not None for entry in plot_data.keys()):
        #Explanation of what is happening here:
        #For plotting two colors the plan is to split up the rectangle at the
        #upwards diagonal and coloring each side with one color
        #Unfortunately there is no direct glyph to do this so we need to use the
        #generic patches method that needs vertices for polygons to draw
        #Here we define a custom Transform that takes the center point of the rectangle
        #and it's size and spits out a list of lists with the coordinates either x or y
        #and for either the upper/lower triangle
        #the strings in the function are the bodies of javascript functions that are inserted into the
        #bokeh framework via CustomJSTransform model
        #x/xs refers to the actual data passed in (defined by bokeh)
        #and all other arguments are defined in arg_dict

        from bokeh.models import CustomJSTransform, Dodge
        from bokeh.transform import transform

        def TriangleTransform(size, data_range, xdata=True, upper=False):
            """Performs a transformation from center points and a block size to triangle coordinates
            to divide a rectangle around this point along the upwards diagonal."""
            #single value transformation
            transform_func = """
                var x_neg = dodge_neg.compute(x)
                var x_pos = dodge_pos.compute(x)
                if (xdata && upper || !xdata && !upper) {
                    res = [x_neg, x_pos, x_neg];
                } else {
                    res = [x_neg, x_pos, x_pos];
                }

                return res
            """
            #vectorized transformation (for array data)
            transform_v_func = """
                const zip= rows=>Array.from(rows[0]).map((_,c)=> rows.map(row=>row[c]));
                var res;
                var x_neg = dodge_neg.v_compute(xs);
                var x_pos = dodge_pos.v_compute(xs);
                if (xdata && upper || !xdata && !upper) {
                    res = zip([x_neg, x_pos, x_neg]);
                } else {
                    res = zip([x_neg, x_pos, x_pos]);
                }
                console.log(res);
                return res;
            """
            arg_dict = {
                'dodge_neg': Dodge(value=-size / 2, range=data_range),
                'dodge_pos': Dodge(value=size / 2, range=data_range),
                'xdata': xdata,
                'upper': upper,
            }

            return CustomJSTransform(func=transform_func, v_func=transform_v_func, args=arg_dict)

        upper = p.patches(transform(entry.x_axis, TriangleTransform(block_size, p.x_range, xdata=True, upper=True)),
                          transform(entry.y_axis, TriangleTransform(block_size, p.y_range, xdata=False, upper=True)),
                          source=source,
                          fill_alpha=0.6,
                          color=plot_params[('color', 0)])
        lower = p.patches(transform(entry.x_axis, TriangleTransform(block_size, p.x_range, xdata=True, upper=False)),
                          transform(entry.y_axis, TriangleTransform(block_size, p.y_range, xdata=False, upper=False)),
                          source=source,
                          fill_alpha=0.6,
                          color=plot_params[('color', 1)])
        r = [upper, lower]
    else:
        r = p.rect(entry.x_axis,
                   entry.y_axis,
                   block_size,
                   block_size,
                   source=source,
                   fill_alpha=0.6,
                   color=plot_params[('color', 0)])
    plot_params.add_tooltips(p, r)

    plot_kw = plot_params.plot_kwargs(plot_type='text', ignore='color')
    # The values displayed on the element boxes
    for (entry, source), kw, position in zip(plot_data.items(), plot_kw, positions):

        p.text(x=dodge(entry.x_axis, x_offset, range=p.x_range),
               y=dodge(entry.y_axis, position, range=p.y_range),
               text=entry.text,
               source=source,
               **kw)

    # add color bar
    if any(entry.color is not None for entry in plot_data.keys()):
        colorbar_options = plot_params['colorbar_options'].copy()
        cbar_fontsize = f"{colorbar_options.pop('fontsize')}pt"

        color_bar = ColorBar(color_mapper=color_mappers[0]['transform'],
                             title_text_font_size='12pt',
                             ticker=BasicTicker(desired_num_ticks=10),
                             border_line_color=None,
                             background_fill_color=None,
                             orientation='vertical',
                             major_label_text_font_size=cbar_fontsize,
                             **colorbar_options)

        p.add_layout(color_bar, 'right')

    # deactivate grid
    p.grid.grid_line_color = None
    plot_params.set_limits(p)
    plot_params.save_plot(p, saveas)

    return p



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module provides the functionality to create/load the schema_dict for the
FleurInputSchema.xsd
"""
from __future__ import annotations

from .fleur_schema_parser_functions import *  #pylint: disable=unused-wildcard-import
from masci_tools.util.xml.common_functions import clear_xml
from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict, CaseInsensitiveFrozenSet
from masci_tools.util.lockable_containers import LockableDict, LockableList
from lxml import etree
import copy
import os
from collections import UserList
from typing import Callable
try:
    from typing import TypedDict, Literal
except ImportError:
    from typing_extensions import TypedDict, Literal  #type:ignore
from . import inpschema_todict


class OutputSchemaData(TypedDict, total=False):
    """Dict representing the entries in the OutputSchemaDict.
       Eventually this should be integrated into the SchemaDict classes
    """
    root_tag: str
    input_tag: str
    iteration_tags: CaseInsensitiveFrozenSet[str]
    inp_version: str
    out_version: str
    tag_paths: CaseInsensitiveDict[str, list[str] | str]
    iteration_tag_paths: CaseInsensitiveDict[str, list[str] | str]
    attrib_types: CaseInsensitiveDict[str, list[AttributeType]]
    text_types: CaseInsensitiveDict[str, list[AttributeType]]
    text_tags: CaseInsensitiveFrozenSet[str]
    unique_attribs: CaseInsensitiveDict[str, str]
    unique_path_attribs: CaseInsensitiveDict[str, list[str]]
    other_attribs: CaseInsensitiveDict[str, list[str]]
    iteration_unique_attribs: CaseInsensitiveDict[str, str]
    iteration_unique_path_attribs: CaseInsensitiveDict[str, list[str]]
    iteration_other_attribs: CaseInsensitiveDict[str, list[str]]
    omitt_contained_tags: LockableList[str]
    tag_info: LockableDict[str, TagInfo]
    iteration_tag_info: LockableDict[str, TagInfo]
    _basic_types: LockableDict[str, list[AttributeType]]
    _input_basic_types: LockableDict[str, list[AttributeType]]


KEYS = Literal['root_tag', 'input_tag', 'iteration_tags', 'tag_paths', 'iteration_tag_paths', '_basic_types',
               'attrib_types', 'text_types', 'text_tags', 'unique_attribs', 'unique_path_attribs', 'other_attribs',
               'iteration_unique_attribs', 'iteration_unique_path_attribs', 'iteration_other_attribs',
               'omitt_contained_tags', 'tag_info', 'iteration_tag_info']


def create_outschema_dict(path: os.PathLike,
                          inpschema_dict: inpschema_todict.InputSchemaData,
                          apply_patches: bool = True) -> OutputSchemaData:
    """
    Creates dictionary with information about the FleurOutputSchema.xsd.
    The functions, whose results are added to the schema_dict and the corresponding keys
    are defined in schema_actions

    :param path: str path to the folder containing the FleurOutputSchema.xsd file
    :param inp_path: str path to the FleurInputSchema.xsd file (defaults to the same folder as path)
    """

    #Add new functionality to this dictionary here
    schema_actions: dict[KEYS, Callable] = {
        'input_tag': get_input_tag,
        'root_tag': get_root_tag,
        'iteration_tags': get_iteration_tags,
        '_basic_types': get_basic_types,
        'attrib_types': extract_attribute_types,
        'text_types': extract_text_types,
        'text_tags': get_text_tags,
        'tag_paths': get_tag_paths,
        'iteration_tag_paths': get_tag_paths,
        'unique_attribs': get_unique_attribs,
        'unique_path_attribs': get_unique_path_attribs,
        'other_attribs': get_other_attribs,
        'iteration_unique_attribs': get_unique_attribs,
        'iteration_unique_path_attribs': get_unique_path_attribs,
        'iteration_other_attribs': get_other_attribs,
        'tag_info': get_tag_info,
        'iteration_tag_info': get_tag_info,
        'omitt_contained_tags': get_omittable_tags,
    }
    schema_patches = [fix_qpoints_typo, patch_text_types]

    xmlschema = etree.parse(path)
    xmlschema, _ = clear_xml(xmlschema)

    xmlschema_evaluator = etree.XPathEvaluator(xmlschema, namespaces=NAMESPACES)
    out_version = eval_single_string_attribute(xmlschema_evaluator, '/xsd:schema/@version')
    out_version_tuple = convert_str_version_number(out_version)

    input_basic_types = inpschema_dict['_basic_types'].get_unlocked()

    schema_dict: OutputSchemaData = {}
    schema_dict['out_version'] = out_version
    for key, action in schema_actions.items():
        addargs: dict[str, dict[str, list[AttributeType]] | bool] = {'input_basic_types': input_basic_types}
        if key in ['unique_attribs', 'unique_path_attribs', 'other_attribs', 'tag_paths', 'tag_info']:
            addargs['stop_iteration'] = True
        elif key in [
                'iteration_unique_attribs', 'iteration_unique_path_attribs', 'iteration_other_attribs',
                'iteration_tag_paths', 'iteration_tag_info'
        ]:
            addargs['iteration_root'] = True
            addargs['iteration'] = True
        schema_dict[key] = action(xmlschema_evaluator, **schema_dict, **addargs)

    schema_dict['_input_basic_types'] = LockableDict(input_basic_types)

    if apply_patches:
        for patch_func in schema_patches:
            patch_func(schema_dict, out_version_tuple)

    return schema_dict


def fix_qpoints_typo(schema_dict: OutputSchemaData, out_version: tuple[int, int]) -> None:
    """
    In versions before 0.35 the attribute qPoints was mistakenly called qpoints
    in the FleurOutputSchema.xsd
    """
    if out_version >= (0, 35):
        #Typo was corrected after this version
        return

    schema_dict['iteration_unique_attribs']['qpoints'] = schema_dict['iteration_unique_attribs']['qpoints'].replace(
        '@qpoints', '@qPoints')

    PATH = './Forcetheorem_DMI'
    old_attribs = set(schema_dict['iteration_tag_info'][PATH]['attribs'].original_case.values())
    old_attribs.discard('qpoints')
    old_attribs.add('qPoints')
    schema_dict['iteration_tag_info'][PATH]['attribs'] = CaseInsensitiveFrozenSet(old_attribs)


def patch_text_types(schema_dict: OutputSchemaData, out_version: tuple[int, int]) -> None:
    """
    Patch the text_types entry to correct ambigouities

    :param schema_dict: dictionary produced by the fleur_schema_parser_functions (modified in-place)
    :param inp_version: input version converted to tuple of ints
    """

    ELEMENTS_ENTRY: Literal['text_types'] = 'text_types'

    if out_version >= (0, 35):
        #After this version the issue was solved
        return

    CHANGE_TYPES = {
        (0, 29): {
            'add': {
                'densityMatrixFor': [AttributeType(base_type='complex', length='unbounded')]
            }
        },
    }

    all_changes: dict[str, list[AttributeType]] = {}

    for version, changes in sorted(CHANGE_TYPES.items(), key=lambda x: x[0]):

        if out_version < version:
            continue

        version_add = changes.get('add', {})
        version_remove = changes.get('remove', set())  #type:ignore

        all_changes = {key: val for key, val in {**all_changes, **version_add}.items() if key not in version_remove}

    for name, new_definition in all_changes.items():
        if name not in schema_dict[ELEMENTS_ENTRY]:
            raise ValueError(f'patch_text_types failed. Type {name} does not exist')
        schema_dict[ELEMENTS_ENTRY][name] = new_definition


def merge_schema_dicts(inputschema_dict: inpschema_todict.InputSchemaData,
                       outputschema_dict: OutputSchemaData) -> OutputSchemaData:
    """
    Merge the information from the input schema into the outputschema
    This combines the type information and adjusts the paths from the inputschema
    to be valid in the out.xml file, i.e. `/fleurInput/cell` becomes `/fleurOutput/fleurInput/cell`
    or `/fleurOutput/inputData/cell` (depending on the version of the output)

    `_basic_types` and `_input_basic_types` stay untouched since they should not be used outside
    of the construction and are merely a informational entry

    :param inputschema_dict: schema dict for the input schema
    :param outputschema_dict: schema dict for the output schema

    :returns: schema dictionary with the information merged
    """
    merged_outschema_dict = copy.deepcopy(outputschema_dict)
    merged_outschema_dict['inp_version'] = inputschema_dict['inp_version']

    input_tag_path = outputschema_dict['tag_paths'][outputschema_dict['input_tag']]
    input_root_tag = inputschema_dict['root_tag']

    #
    # 1. Merge path entries and modify the input paths
    #
    path_entries: set[Literal['tag_paths', 'unique_attribs', 'unique_path_attribs', 'other_attribs']] = {
        'tag_paths', 'unique_attribs', 'unique_path_attribs', 'other_attribs'
    }
    for entry in path_entries:
        for key, val in inputschema_dict[entry].items():

            paths: list[str] | str = merged_outschema_dict[entry].get(key, UserList())

            if not isinstance(paths, (UserList, list)):
                paths = [paths]
            paths_set = set(paths)

            if not isinstance(val, (UserList, list)):
                val = [val]
            new_paths_set = {f"{input_tag_path}{inp_path.replace(f'/{input_root_tag}','')}" for inp_path in val}

            new_paths = sorted(paths_set.union(new_paths_set))
            if len(new_paths) == 1:
                merged_outschema_dict[entry][key] = new_paths[0]
            else:
                merged_outschema_dict[entry][key] = new_paths

    #Remove the root_tag of the input if it is different from the input tag in the out.xml
    if input_root_tag != outputschema_dict['input_tag']:
        merged_outschema_dict['tag_paths'].pop(input_root_tag)

    #
    # 2. Insert tag_info paths (There is no possibility for overlap here)
    #
    new_tag_info_entries = {
        f"{input_tag_path}{path.replace(f'/{input_root_tag}','')}": info
        for path, info in inputschema_dict['tag_info'].items()
    }
    merged_outschema_dict['tag_info'].update(new_tag_info_entries)

    #
    # 3. Merge the attribute type information
    #
    for attrib, types in inputschema_dict['attrib_types'].items():
        out_types = set(merged_outschema_dict['attrib_types'].get(attrib, []))

        new_types = sorted(out_types.union(set(types)), key=type_order)

        merged_outschema_dict['attrib_types'][attrib] = new_types

    #
    # 4. Merge the definition information for text tags
    #
    for tag, types in inputschema_dict['text_types'].items():
        out_types = set(merged_outschema_dict['text_types'].get(tag, []))

        new_types = sorted(out_types.union(set(types)), key=type_order)

        merged_outschema_dict['text_types'][tag] = new_types

    merged_outschema_dict['text_tags'] = merged_outschema_dict['text_tags'] | inputschema_dict['text_tags']

    #
    # 5. Merge the omittable tags
    #
    merged_outschema_dict['omitt_contained_tags'] = LockableList(
        sorted(set(merged_outschema_dict['omitt_contained_tags']).union(inputschema_dict['omitt_contained_tags'])))

    return merged_outschema_dict
